from itertools import product
from overrides import overrides

import pytorch_lightning as pl
from sklearn.base import BaseEstimator
from sklearn.model_selection import StratifiedKFold, train_test_split

from typing import Union, List, Dict, Any, Optional
from astrape.utilities.utils_lightning import *
from astrape.base.experiment_base import BaseExperiment
from astrape.constants.astrape_constants import *
from astrape.exceptions.exceptions import *


class CrossValidation(BaseExperiment):
    r"""Performs stratified K-fold cross validation.

    Attributes:

        rng: Random number generator.

        seed_num (int): Random seed number. 

        trainer_config (dict): Trainer configurations.

        search_type (str): "grid" or "random".

        parameters (dict): Dictionary specifying sets of hyperparameters.

        val_metric (str): The validation metric.

        cv (int): Number of folds. i.e., K in stratified K-fold cross validation.

        jsonlog_dir (str): Path where json files will be logged.

        ckpt_dir (str): Path where model checkpoints will be logged. 

        best_val_score (float): The best validation score.

        n_random_search (int): Number of sampling from distribution when performing random search. 
    """

    def __init__(
        self,
        project_name : str,
        dims : int,
        X,
        y,
        n_classes : int,
        X_test,
        y_test,
        test_size : float,
        stack_models : bool,
        path : str,
        model_type : Union["BaseEstimator", "pl.LightningModule"],
        parameters : Dict[str, List[Any]],
        trainer_config : Optional[Dict],
        val_metric : str ='val/acc',
        mode : Optional[str] = None,
        cv : int = 5,
        search_type : str = 'random',
        n_random_search : Optional[int] = None,
        random_number : int = DEFAULT_RANDOM_SEED_NUMBER
    ):

        """

        Args:

            project_name (str): Name of the project. Directory of {path}/{project_name} will be created.

            X (np.ndarray or torch.Tensor or pd.DataFrame): Data for the project. If X_test is not specified, X would be splitted(stratify=y) into train/val data with a given random state.

            y (np.ndarray or torch.Tensor or pd.DataFrame): Labels for the project. If y_test is not specified, y would be splitted(stratify=y) into train/val labels with a given random state.

            X_train (torch.Tensor): Training data.   

            y_train (torch.Tensor): Training labels.

            X_val (np.ndarray or torch.Tensor or pd.DataFrame, optional): Validation data if explicitly exists.   

            y_val (np.ndarray or torch.Tensor or pd.DataFrame, optional): Validation labels if explicitly exists.

            X_test (np.ndarray or torch.Tensor or pd.DataFrame, optional): Test data if explicitly exists.
            
            y_test (np.ndarray or torch.Tensor or pd.DataFrame, optional): Test labels if explicitly exists.

            test_size (float): Test(and validation) size when splitting the data.
                
            n_classes (int): Number of classes for the classification task. Don't specify this when performing regression tasks.

            path (str): Base path for the project. Default is the current directory.

            model_type (`LightningModule`, `BaseEstimator): The type of model (class)

            parameters (dict): Dictionary specifying sets of hyperparameters.

            trainer_config (dict): Trainer configurations. 
                Default: ``None``
            
            val_metric (str): The validation metric.
                Default: ``'val/acc'``

            mode (str): Mode for validation metric. Either "min" or "max". If left as None, mode will automatically be set for 'val/acc', 'val/auc', and 'val/loss'. If the metric is not one of these, error will be raised.
                Default: ``None``

            cv (int): K in K-fold cross validation.
                Default: ``5``

            search_type (str): Search type of the cross validation. Should be "random" or "grid".
                Default: ``'random'``

            n_random_search (int, optional): Number of sampling from distribution when performing random search. If None, will be set to 5.
                Default: ``None``

            random_number (int): Random number for random number generator.
                Default: ``2022``

        Example::
            
            >>> parameters= { 
                                "n_layer_1": {
                                # Choose from pre-defined values
                                "values": [32, 64, 128, 256, 512]
                                },
                                "n_layer_2": {
                                # Choose from pre-defined values
                                "values": [32, 64, 128, 256, 512, 1024]
                                },
                                "lr": {
                                    # log uniform distribution between exp(min) and exp(max)
                                    "distribution": "log_scale_uniform",
                                    "min": -9.21,   # exp(-9.21) = 1e-4
                                    "max": -4.61    # exp(-4.61) = 1e-2
                                }
                            }
        Returns:

            None
        """
        self.rng = np.random.default_rng(random_number)
        self.seed_num = int(10**6 * self.rng.random(1))
        
        super().__init__(
            project_name=project_name,
            X=X,
            y=y,
            n_classes=n_classes,
            X_test=X_test,
            y_test=y_test,
            test_size=test_size,
            stack_models=stack_models,
            random_number=self.seed_num,
            path=path
        )
        self.model_type = model_type
        self.trainer_config = trainer_config if trainer_config else {'max_epochs':2}
        self.search_type = search_type
        self.parameters = parameters
        self.val_metric = val_metric
        self.cv = cv
        self.dims = dims

        if 'batch_size' not in self.parameters.keys():
            self.parameters.update({'batch_size' : {'values' : [DEFAULT_BATCH_SIZE]}})
        self.jsonlog_dir = None
        self.ckpt_dir = None
        if not mode:
            if val_metric == 'val/loss':
                self.mode = 'min'
            elif val_metric in ['val/acc', 'val/auc']:
                self.mode = 'max'
            else:
                raise AssertionError("You should specify the mode if val_metric is not one of 'val/acc', 'val/loss', or 'val/auc'.")
        self.best_val_score = None
        self.n_random_search = 5 if not n_random_search else n_random_search

    def get_cv_indices(self):
        skf = StratifiedKFold(n_splits=self.cv, random_state=self.random_state, shuffle=True)
        cv_indices = []
        for train_idx, val_idx in skf.split(self.X, self.y):
            cv_indices.append((train_idx, val_idx))

        return cv_indices


    def get_hparam_info(self)->Dict:
        r"""Get informations of ``parameters``.

        Returns:

            dict: Information of ``parameters``.
        """
        hparam_info = {}
        for hparam_name, hparam_config in self.parameters.items():
            if 'values' in hparam_config.keys():
                value_list = list(hparam_config.values())[0]
                hparam_info.update({hparam_name : value_list})
                
            elif 'distribution' in hparam_config.keys():
                value_info = {hparam_config['distribution'] : (hparam_config['min'], hparam_config['max'])}
                hparam_info.update({hparam_name : value_info})
            else:
                raise ValueError("The scheme of the values for the hyperparameters should either be 'values' or 'distribution'.")

        return hparam_info

    def get_hparams_dict_list(self)->List[Dict]:
        r"""Get set of hyperparameter candidates.

        Returns:

            list: List of hyperparametr candidates.
        """
        hparams_dict_list = []
        hparam_info = self.get_hparam_info()
        if self.search_type == 'random':
            for _ in range(self.n_random_search):
                hparams_dict = {}
                for hparam_name, value_info in hparam_info.items():
                    if isinstance(value_info, list):
                        idx = np.random.randint(low=0, high=len(value_info))
                        hparam_value = value_info[idx]
                    else: # distribution
                        for key, item in value_info.items():
                            dist_type = key
                            dist_config = item
                        if dist_type == 'uniform':
                            hparam_value = np.random.uniform(*dist_config)
                        elif dist_type == 'log_scale_uniform':
                            power = np.random.uniform(*dist_config)
                            hparam_value = 10 ** power
                    hparams_dict.update({hparam_name : hparam_value})
                hparams_dict_list.append(hparams_dict)
        elif self.search_type == 'grid':
            grid_dict = {}
            hparams_dict = {}
            for hparam_name, value_info in hparam_info.items():
                if isinstance(value_info, dict): # pick one random sample from the RV
                    for key, item in value_info.items():
                        dist_type = key
                        dist_config = item
                    if dist_type == 'uniform':
                        hparam_value = np.random.uniform(*dist_config)
                    elif dist_type == 'log_scale_uniform':
                        power = np.random.uniform(*dist_config)
                        hparam_value = 10 ** power
                    hparams_dict.update({hparam_name : hparam_value})
                else: # grid list
                    grid_dict.update({hparam_name : value_info}) 
            grid_hparam_names, grid_hparam_values = zip(*grid_dict.items())
            for bundle in product(*grid_hparam_values):
                hparams_dict = {}
                grid_sample = dict(zip(grid_hparam_names, bundle))
                hparams_dict.update(grid_sample)
                hparams_dict_list.append(hparams_dict)
        return hparams_dict_list

    def get_data_module_ms(
        self,
        X : "np.ndarray",
        y : "np.ndarray",
        batch_size : int,
        X_val : Optional["np.ndarray"] = None,
        y_val : Optional["np.ndarray"] = None,
        X_test : Optional["np.ndarray"] = None,
        y_test : Optional["np.ndarray"] = None,
        test_size : float = 0.01
    )->Union["pl.LightningDataModule", Tuple["np.ndarray","np.ndarray","np.ndarray","np.ndarray","np.ndarray","np.ndarray"]]:
        r"""Gets datamodule.
        
        Returns:

            :obj:`LightningDataModule` or tuple: The datamodule. Returns tuple of numpy ndarrays when using sklearn models.
        """
        if BaseEstimator in inspect.getmro(self.model_type):
            if (X_val or y_val) and not (X_val and y_val):
                raise ValueError("X_val and y_val should have the same # of samples.")
            if (X_test or y_test) and not (X_test and y_test):
                raise ValueError("X_test and y_test should have the same # of samples.")
            elif X_val and y_val and X_test and y_test:
                data_module = (X, X_val, y, y_val, X_test, y_test)
            elif X_test and y_test:
                X_train, X_val, y_train, y_val = train_test_split(X, y, test_size=test_size, random_state=self.random_state, stratify=y)
                data_module = (X_train, X_val, y_train, y_val, X_test, y_test)
            elif X_val and y_val:
                X_train, X_test, y_train, y_test = train_test_split(X, y, test_size=test_size, random_state=self.random_state, stratify=y)
                X_train, X_val, y_train, y_val = train_test_split(X, y, test_size=test_size, random_state=self.random_state, stratify=y)
                data_module = (X_train, X_val, y_train, y_val, X_test, y_test)
            elif X_val and y_val:
                X_train, X_test, y_train, y_test = train_test_split(X, y, test_size=test_size, random_state=self.random_state, stratify=y)
                data_module = (X_train, X_val, y_train, y_val, X_test, y_test)

        elif pl.LightningModule in inspect.getmro(self.model_type):
            data_module = initialize_datamodule(
                batch_size=batch_size, 
                X=X, 
                y=y,
                X_val=X_val,
                y_val=y_val,
                X_test=X_test,
                y_test=y_test,
                test_size=test_size,
                random_state=self.random_state
            )

        return data_module

    def search(self)->Tuple[Union["pl.LightningModule", "BaseEstimator"], Dict]:
        r"""Performs cross-validation.

        Returns:

            tuple: The best model structure and its information.
        """
        self.fit_or_cv = "CV"
        self.update_log_path()
        cv_indices = self.get_cv_indices()

        best_val_score = None
        best_hparams = None
        hparams_dict_list = self.get_hparams_dict_list()
        for hparam_idx, hparams_dict in enumerate(hparams_dict_list):
            print("Stratified {}-Fold validation with {} set of hyperparameters. Total {} runs will be conducted.".format(self.cv, len(hparams_dict_list), self.cv*len(hparams_dict_list)))
            val_score_per_run = []
            for cv_idx, (train_idx, val_idx) in enumerate(cv_indices):
                print("trying {}/{} set of hyperparameters.".format(hparam_idx+1, len(hparams_dict_list)))
                print("running {}/{} folds.".format(cv_idx+1, self.cv))
                X_train, X_val = self.X[train_idx], self.X[val_idx]
                y_train, y_val = self.y[train_idx], self.y[val_idx]
                self.model = self.set_model(
                            model_type=self.model_type,
                            **hparams_dict
                        )
                self.jsonlog_dir = self.log_path + "/logs/jsonlogs/" + self.model_metadata
                self.ckpt_dir = self.log_path + "/logs/checkpoints/" + self.model_metadata
                
                if isinstance(self.model, pl.LightningModule):
                    ckpt_extension = ".ckpt"
                elif isinstance(self.model, BaseEstimator):
                    ckpt_extension = ".sav"
                self.ckpt_dir = self.ckpt_dir + ckpt_extension

                self.data_module = self.get_data_module_ms(
                    X=X_train,
                    y=y_train,
                    batch_size=hparams_dict['batch_size'],
                    X_val=X_val,
                    y_val=y_val,
                    X_test=self.X_test,
                    y_test=self.y_test,
                    test_size=self.test_size
                )
                
                self.trainer = self.set_trainer(**(self.trainer_config))
            
                self.fit(fit_or_cv="CV",**self.trainer_config)
                self.save_ckpt()
                val_score = self.trainer.validate(self.model, self.data_module)[0][self.val_metric]
                val_score_per_run.append(val_score)

            val_score_per_run = np.array(val_score_per_run)
            avg_val_score_in_run = val_score_per_run.mean()
            
            if not best_val_score:
                best_val_score = avg_val_score_in_run
                best_hparams = hparams_dict
            else:
                if avg_val_score_in_run < best_val_score and self.mode == "min":
                    best_val_score = avg_val_score_in_run
                    best_hparams = hparams_dict   
                elif avg_val_score_in_run > best_val_score and self.mode == "max":
                    best_val_score = avg_val_score_in_run
                    best_hparams = hparams_dict
                    
        self.best_model_structure = self.best_ckpt_thus_far(val_metric=self.val_metric)[0]
        
        self.best_hparams = best_hparams
        self.best_val_score = best_val_score
        info = {
            'best_hparams' : self.best_hparams,
            self.val_metric : self.best_val_score
        }
        self.fit_or_cv = "FIT"
        return (self.best_model_structure, info)                    



APPLICABLE_OPTIMIZER_TYPE_LIST = ["adam", "sgd"]
APPLICABLE_VAL_METRIC_LIST = ["val/loss", "val/acc"]

DEFAULT_BATCH_SIZE = 256
DEFAULT_LR = 1e-4
DEFAULT_N_LAYERS = 2
DEFAULT_DROPOUT_P = 0
DEFAULT_BN = True
DEFAULT_L1_STRENGTH = 0
DEFAULT_WEIGHT_DECAY = 0
DEFAULT_OPTIMIZER_TYPE = "adam"

DEFAULT_RANDOM_SEED_NUMBER = 2022

DEFAULT_MAX_EPOCHS = 100
DEFAULT_AUTO_LR_FIND = False



from astrape.constants.astrape_constants import APPLICABLE_OPTIMIZER_TYPE_LIST, APPLICABLE_VAL_METRIC_LIST

from astrape.constants.astrape_constants import DEFAULT_BATCH_SIZE, DEFAULT_LR, DEFAULT_N_LAYERS, DEFAULT_DROPOUT_P, DEFAULT_BN, DEFAULT_L1_STRENGTH
from astrape.constants.astrape_constants import DEFAULT_WEIGHT_DECAY, DEFAULT_OPTIMIZER_TYPE

from astrape.constants.astrape_constants import DEFAULT_RANDOM_SEED_NUMBER

from astrape.constants.astrape_constants import DEFAULT_MAX_EPOCHS, DEFAULT_AUTO_LR_FIND

__all__ = [
    APPLICABLE_OPTIMIZER_TYPE_LIST ,
    APPLICABLE_VAL_METRIC_LIST ,
    DEFAULT_BATCH_SIZE ,
    DEFAULT_LR ,
    DEFAULT_N_LAYERS ,
    DEFAULT_DROPOUT_P ,
    DEFAULT_BN ,
    DEFAULT_L1_STRENGTH ,
    DEFAULT_WEIGHT_DECAY ,
    DEFAULT_OPTIMIZER_TYPE ,
    DEFAULT_RANDOM_SEED_NUMBER ,
    DEFAULT_MAX_EPOCHS,
    DEFAULT_AUTO_LR_FIND 
]


from astrape.models.models_lightning import MLP, ContractingMLP, CustomMLP, VGG, UNet
from astrape.models.model_buildingblocks import DoubleConv, Down, Up
from astrape.utilities.utils_lightning import set_default_parameters
__all__ = [
    'MLP', 
    'ContractingMLP', 
    'CustomMLP', 
    'VGG', 
    'UNet', 
    'DoubleConv', 
    'Down', 
    'Up', 
    'set_default_parameters'
    ]



import pytorch_lightning as pl 
from typing import Union, List, Dict, Any, Optional, Tuple, cast
from astrape.utilities.dataloader_lightning import *
import inspect
from astrape.constants.astrape_constants import *
from sklearn.base import BaseEstimator

def set_default_parameters(config)->None:
    r"""Sets default parameters for astrape models.
    """
    if 'dropout_p' not in config.keys():
        config.update({'dropout_p' : DEFAULT_DROPOUT_P})
    if 'bn' not in config.keys():
        config.update({'bn' : DEFAULT_BN})
    if 'lr' not in config.keys():
        config.update({'lr' : DEFAULT_LR})
    if 'weight_decay' not in config.keys():
        config.update({'weight_decay' : DEFAULT_WEIGHT_DECAY})
    if 'optimizer_type' not in config.keys():
        config.update({'optimizer_type' : DEFAULT_OPTIMIZER_TYPE})
    if 'batch_size' not in config.keys():
        config.update({'batch_size' : DEFAULT_BATCH_SIZE})
    if 'l1_strength' not in config.keys():
        config.update({'l1_strength' : DEFAULT_L1_STRENGTH})

def initialize_model(
    model_type : "pl.LightningModule", 
    dims : int, 
    n_classes : int, 
    **config : Dict
)->Union["pl.LightningModule", "BaseEstimator"]:
    r"""Initializes Lightning model.
    """
    if BaseEstimator in inspect.getmro(model_type):
        model = model_type(**config)
    else:
        model = model_type(dims=dims, n_classes=n_classes, **config)

    return model
    
def initialize_datamodule(
    **config
)->Union["pl.LightningDataModule", Tuple]:
    r"""

    Args:

        **config (Any): Dictionary of the followings
            batch_size 
            X 
            y 
            X_test 
            y_test 
            test_size 
            random_state
    """
    data_module = DataModule(**config)
    return data_module




from astrape.utilities.dataloader_lightning import DataModule
from astrape.utilities.utils_lightning import initialize_datamodule, initialize_model
from astrape.utilities.utils_lightning import set_default_parameters
__all__ = [
    "DataModule", 
    "initialize_datamodule", 
    "initialize_model", 
    "set_default_parameters"
    ]



from pickle import BINBYTES
from random import sample
import torch
import torch.nn as nn
from astrape.utilities.utils_lightning import set_default_parameters
from astrape.base.model_base import BaseNet, ConvNet
from typing import Union, List, Dict, Any, Optional, cast, Tuple
from astrape.exceptions.exceptions import *
from astrape.models.model_buildingblocks import Down, DoubleConv, Up
import math
import numpy as np



class _MLP(nn.Module):
    r"""Multilayer perceptron with identical # of hidden units.

    Args:

        dims (int or tuple of int): The dimension of the data. e.g., 4, (8,8,1)

        n_classes (int, optional): Number of classes (labels) in the classification task.
        
        n_layers (int): Number of layers. If 1, it would be an old-school perceptron i.e., logistic regression.
            Default: ``2``

        n_hidden_units (int): Number of hidden units for all hidden layers. 
            Default : ``10`` 

        batch_size (int): Batch size. 
            Default : ``64``

        optimizer_type (str): Type of the optimizer. ("adam" or "sgd") 
            Default : ``"adam"`` 

        dropout_p (float): Dropout probability 
            Default : ``0``

        bn (bool): Whether to use batch normalization or not. 
            Default : ``True``

        weight_decay (float): Coefficient for L2-regularization. 
            Default : ``0`` 

        lr (float): Learning rate. 
            Default: ``1e-4``
    """
    def __init__(
        self,
        dims : Union[int, Tuple[int]],
        n_classes : int, 
        n_layers : int,
        n_hidden_units : int,
        batch_size : int,
        optimizer_type : str,
        dropout_p : float,
        bn : bool,
        lr : float,
        weight_decay : float,
        l1_strength : float,
        activation_type : str = "relu", # "relu" or "leaky_relu"
        negative_slope : Optional[float] = None, # negative slope for leaky relu   
        sample_weight : Optional[Union["np.ndarray", "torch.Tensor"]] = None
    ) -> None:

        super(_MLP, self).__init__()
        layers = []
        
        if isinstance(dims, tuple):
            flattened = 1
            for dim in dims:
                flattened *= dim
            dims = flattened 

        self.dims = dims
        self.n_classes = n_classes
        self.n_layers = n_layers
        self.n_hidden_units = n_hidden_units
        self.optimizer_type = optimizer_type
        self.batch_size = batch_size
        self.dropout_p = dropout_p
        self.bn = bn
        self.lr = lr
        self.weight_decay = weight_decay 
        self.sample_weight = sample_weight
        self.l1_strength = l1_strength
        self.activation_type = activation_type
        self.negative_slope = negative_slope
        if self.n_layers == 1:
            if self.n_classes > 2:
                layers.append(nn.Linear(self.dims, self.n_classes))
            else:
                layers.append(nn.Linear(self.dims, 1))

        else : #Deep  Neural Network
            fc1 = nn.Linear(self.dims, self.n_hidden_units)
            layers.append(fc1)
            if self.bn:
                layers.append(nn.BatchNorm1d(num_features=self.n_hidden_units))
            
            if activation_type == "tanh":
                layers.append(nn.Tanh())
            elif activation_type == "relu":
                layers.append(nn.ReLU())
            elif activation_type == "leaky_relu":
                if not negative_slope:
                    negative_slope = 1e-2
                layers.append(nn.LeakyReLU(negative_slope=negative_slope))
            else:
                raise ValueError(f"Wrong name {activation_type} for activation function.")
            layers.append(nn.Dropout(self.dropout_p))
            
            for i in range(self.n_layers-2):
                fc2 = nn.Linear(self.n_hidden_units, self.n_hidden_units)
                layers.append(fc2)
                if self.bn:
                    layers.append(nn.BatchNorm1d(num_features=self.n_hidden_units))
                
                if activation_type == "tanh":
                    layers.append(nn.Tanh())
                elif activation_type == "relu":
                    layers.append(nn.ReLU())
                elif activation_type == "leaky_relu":
                    if not negative_slope:
                        negative_slope = 1e-2
                    layers.append(nn.LeakyReLU(negative_slope=negative_slope))
                else:
                    raise ValueError(f"Wrong name {activation_type} for activation function.")
                layers.append(nn.Dropout(self.dropout_p))

            
            if self.n_classes > 2:
                fc3 = nn.Linear(self.n_hidden_units, self.n_classes)
                layers.append(fc3)
            else: # including the case when self.n_classes is None
                fc3 = nn.Linear(self.n_hidden_units, 1)
                layers.append(fc3)
        
        if self.n_classes is None:
            pass
        elif self.n_classes> 2:
            layers.append(nn.Softmax(dim=1))
        else:
            layers.append(nn.Sigmoid())

    
        self.layers = nn.Sequential(*layers)
    
    def forward(self, X):
        return self.layers(X)


class MLP(BaseNet):
    r"""Multi-layer perceptron with identical # of hidden units.
    """
    def __init__(
        self,
        dims : Union[int, Tuple[int]],
        n_classes : int,
        sample_weight : Optional[Union["torch.Tensor", "np.ndarray"]] = None,
        **hparams : Any
    ) -> None:
        r"""Multi-layer perceptron with identical # of hidden units.

            The number of layers are specified with argument ``n_layers``.

            Args:

                dims (int or tuple of int): The dimension of the data. e.g., 4, (8,8,1), (32,32,3)

                n_classes (int, optional): Number of classes (labels) in the classification task.
                
                n_layers (int): Number of layers. If 1, it would be an old-school perceptron i.e., logistic regression.
                    Default: ``2``

                n_hidden_units (int): Number of hidden units for all hidden layers. 
                    Default : ``10`` 

                batch_size (int): Batch size. 
                    Default : ``64``

                optimizer_type (str): Type of the optimizer. ("adam" or "sgd") 
                    Default : ``"adam"`` 

                dropout_p (float): Dropout probability 
                    Default : ``0``

                bn (bool): Whether to use batch normalization or not. 
                    Default : ``True``

                weight_decay (float): Coefficient for L2-regularization. 
                    Default : ``0`` 

                lr (float): Learning rate. 
                    Default: ``1e-4``
        """
        
        entire_keys = [
            'dims', 
            'n_classes', 
            'n_layers',
            'n_hidden_units', 
            'optimizer_type', 
            'batch_size',
            'dropout_p',
            'bn',
            'sample_weight',
            'lr',
            'weight_decay',
            'l1_strength',
            'activation_type',
            'negative_slope'
            ]
        default_parameter_dict = {
            'n_layers' : 2,
            'n_hidden_units' : 10
        }
        
        self.dims = dims
        self.n_classes = n_classes
        self.sample_weight = sample_weight
        
        # set default for model-specific hyperparameters
        for argument, default in default_parameter_dict.items():
            if argument not in hparams.keys(): 
                hparams.update({argument : default})
        # set default for model-agnostic hyperparameters (read utils_lightning.py)
        set_default_parameters(hparams)
        # save information of **hparams to self.hparams.
        self.save_hyperparameters(*hparams)
        # validate contraction_ratio
        
        super(MLP, self).__init__(
            dims=dims, 
            n_classes=n_classes, 
            sample_weight=sample_weight,
            **hparams
        )
        # sanity check. Check if right arguments are declared.
        for key in hparams.keys():
            if key not in entire_keys:
                raise ValueError(f'Wrong argument {key} is passed to the model.')
        self.model = _MLP(dims=self.dims, n_classes=self.n_classes, sample_weight=self.sample_weight, **hparams)

    def forward(self, X):
        return self.model.forward(X)
########################################################## _MLP & MLP (Ends) ##########################################################


########################################################## _ContractingMLP & ContractingMLP (Begins) ##########################################################

class _ContractingMLP(nn.Module):
    r"""ContractingMLP with given contraction ratio.

    _ContractingMLP and ContractingMLP defines an MLP with number of hidden units decreasing with a factor
    of contraction_ratio.
    Therefore, the hyperparameter that is additionally defined in ContractingMLP is contraction_ratio.

    _ContractingMLP and ContractingMLP are basically identical in structure. The only difference is that
    _ContractingMLP is an offspring of nn.Module and ContractingMLP is an offspring of BaseNet.
    """
    def __init__(
        self,
        dims : Union[int, Tuple[int]],
        n_classes : int,
        contraction_ratio : float, 
        optimizer_type : str, 
        batch_size : int ,
        dropout_p : float,
        bn : bool,
        lr : float,
        weight_decay : float,
        l1_strength : float,
        activation_type : str = "relu",
        negative_slope : Optional[float] = None,
        sample_weight : Optional[Union["torch.Tensor", "np.ndarray"]] = None
    ) -> None:
        r"""Initialize ContractingMLP
        
            Args:
            
                dims (int or tuple of int): The dimension of the data. e.g., 4, (8,8,1), (32,32,3)

                n_classes (int, optional): Number of classes (labels) in the classification task.
                
                contraction_ratio (float): The 0<contraction ratio<1 of the model. Number of hidden units contracts with this ratio.
                    Default: ``0.5``

                batch_size (int): Batch size. 
                    Default : ``64``

                optimizer_type (str): Type of the optimizer. ("adam" or "sgd") 
                    Default : ``"adam"`` 

                dropout_p (float): Dropout probability 
                    Default : ``0``

                bn (bool): Whether to use batch normalization or not. 
                    Default : ``True``

                weight_decay (float): Coefficient for L2-regularization. 
                    Default : ``0`` 

                lr (float): Learning rate. 
                    Default: ``1e-4``
    """
        super(_ContractingMLP, self).__init__()

        if isinstance(dims, tuple):
            flattened = 1
            for dim in dims:
                flattened *= dim
            dims = flattened 

        self.dims = dims
        self.n_classes = n_classes
        self.contraction_ratio = contraction_ratio
        self.optimizer_type = optimizer_type
        self.batch_size = batch_size
        self.dropout_p = dropout_p
        self.bn = bn
        self.lr = lr
        self.weight_decay = weight_decay  
        self.sample_weight = sample_weight
        self.l1_strength = l1_strength
        self.current_units = self.dims
        self.activation_type = activation_type
        self.negative_slope = negative_slope
        
        
        layers = []

        if int(self.current_units * self.contraction_ratio) <= self.n_classes:
            if self.n_classes > 2:
                layers.append(nn.Linear(self.dims, self.n_classes))
            else:
                layers.append(nn.Linear(self.dims, 1))

        else : #Deep  Neural Network
            self.current_units = int(self.current_units * self.contraction_ratio)
            fc1 = nn.Linear(self.dims, self.current_units)
            layers.append(fc1)
            if self.bn:
                layers.append(nn.BatchNorm1d(num_features=self.current_units))
            if activation_type == "tanh":
                layers.append(nn.Tanh())
            elif activation_type == "relu":
                layers.append(nn.ReLU())
            elif activation_type == "leaky_relu":
                if not negative_slope:
                    negative_slope = 1e-2
                layers.append(nn.LeakyReLU(negative_slope=negative_slope))
            else:
                raise ValueError(f"Wrong name {activation_type} for activation function.")
            
            layers.append(nn.Dropout(self.dropout_p))

            
            while int(self.current_units * self.contraction_ratio) >= self.n_classes:

                
                fc2 = nn.Linear(self.current_units,  int(self.current_units * self.contraction_ratio))
                self.current_units = int(self.current_units * self.contraction_ratio)
                layers.append(fc2)
                if self.bn:
                    layers.append(nn.BatchNorm1d(num_features=self.current_units))

                if activation_type == "tanh":
                    layers.append(nn.Tanh())
                elif activation_type == "relu":
                    layers.append(nn.ReLU())
                elif activation_type == "leaky_relu":
                    if not negative_slope:
                        negative_slope = 1e-2
                    layers.append(nn.LeakyReLU(negative_slope=negative_slope))
                else:
                    raise ValueError(f"Wrong name {activation_type} for activation function.")
                
                layers.append(nn.Dropout(self.dropout_p))
                
            if self.n_classes > 2:
        
                fc3 = nn.Linear(self.current_units, self.n_classes)
                layers.append(fc3)
            else:
                fc3 = nn.Linear(self.current_units, 1)
                layers.append(fc3)
        
        if self.n_classes> 2:
            layers.append(nn.Softmax(dim=1))
        else:
            layers.append(nn.Sigmoid())


        self.layers = nn.Sequential(*layers)


    def forward(self, X):
        return self.layers(X)

class ContractingMLP(BaseNet):
    r"""ContractingMLP with given contraction ratio.
    
    _ContractingMLP and ContractingMLP defines an MLP with number of hidden units decreasing with a factor
    of contraction_ratio.
    Therefore, the hyperparameter that is additionally defined in ContractingMLP is contraction_ratio.

    _ContractingMLP and ContractingMLP are basically identical in structure. The only difference is that
    _ContractingMLP is an offspring of nn.Module and ContractingMLP is an offspring of BaseNet.
    """
    def __init__(
        self,
        dims : Union[int, Tuple[int]],
        n_classes : int,
        sample_weight : Optional[Union["torch.Tensor", "np.ndarray"]] = None,
        **hparams : Any
    ) -> None:
        r"""Initialize ContractingMLP
        
            Args:
            
                dims (int or tuple of int): The dimension of the data. e.g., 4, (8,8,1), (32,32,3)

                n_classes (int, optional): Number of classes (labels) in the classification task.
                
                contraction_ratio (float): The 0<contraction ratio<1 of the model. Number of hidden units contracts with this ratio.
                    Default: ``0.5``

                batch_size (int): Batch size. 
                    Default : ``64``

                optimizer_type (str): Type of the optimizer. ("adam" or "sgd") 
                    Default : ``"adam"`` 

                dropout_p (float): Dropout probability 
                    Default : ``0``

                bn (bool): Whether to use batch normalization or not. 
                    Default : ``True``

                weight_decay (float): Coefficient for L2-regularization. 
                    Default : ``0`` 

                lr (float): Learning rate. 
                    Default: ``1e-4``
        """
        
        entire_keys = [
            'dims', 
            'n_classes', 
            'sample_weight',
            'contraction_ratio', 
            'optimizer_type', 
            'batch_size',
            'dropout_p',
            'bn',
            'lr',
            'weight_decay',
            'l1_strength',
            'activation_type',
            'negative_slope'
            ]
        default_parameter_dict = {
            'contraction_ratio' : 0.5
        }
        
        # set default for model-specific hyperparameters
        for argument, default in default_parameter_dict.items():
            if argument not in hparams.keys(): 
                hparams.update({argument : default})
        #set default for model-agnostic hyperparameters
        set_default_parameters(hparams)
        # save information of **hparams to self.hparams.
        self.save_hyperparameters(*hparams)
        # validate contraction_ratio
        if not self.hparams.contraction_ratio < 1 and not self.hparams.contraction_ratio >0 :
            raise ValueError("'contraction_ratio' must be values between 0 and 1") 

        super(ContractingMLP, self).__init__(
            dims=dims,
            n_classes=n_classes,
            sample_weight=sample_weight,
            **hparams
        )

        # sanity check. Check if right arguments are declared.
        for key in hparams.keys():
            if key not in entire_keys:
                raise ValueError('Wrong argument is passed to the model.')
        self.model = _ContractingMLP(dims=dims, n_classes=n_classes, sample_weight=sample_weight, **hparams)

    def forward(self, X):
        return self.model.forward(X)

########################################################## _ContractingMLP & ContractingMLP (Ends) ##########################################################

########################################################## _CustomMLP & CustomMLP (Begins) ##########################################################
"""
_CustomMLP and CustomMLP defines an MLP with number of hidden units for hidden layers 
customized for all layers. 

We can define n_hidden_units for all hidden layers with a list of integers.

The hyperparameters that are only in CustomMLP are 'layer_{n}_dim's (for all hidden layers). 


_CustomMLP and CustomMLP are basically identical in structure. The only difference is that
_CustomMLP is an offspring of nn.Module and CustomMLP is an offspring of BaseNet.

Arguments:
    - dims (Union[int, Tuple[int]]): The dimension of the data. e.g., 4, (8,8,1)

    - n_classes (Optional[int]) : Number of classes (labels) in the classification task.
                                    Do not specify this argument (set is as None) when you 
                                re performing regression tasks.
    
    - n_hidden_units_list : Optional[List[int]] : List of all n_hidden_units for all hidden layers.
                                                  If not specified, it would be an old-school perceptron.

    - batch_size (int) : Batch size. Default : 64 (see utils_lightning.py)

    - optimizer_type (str) : Type of the optimizer. ("adam" or "sgd") Default : "adam" (see utils_lightning.py) 
                            #TODO : implement other optimizers.(in model_base.py)

    - dropout_p (0<= float < 1) : Dropout probability Default : 0 (see utils_lightning.py)

    - bn (bool) : Whether to use batch normalization or not. Default : True (see utils_lightning.py)

    - weight_decay (0<= float) : coefficient for L2-regularization. Default : 0 (see utils_lightning.py)

    - lr ( 0 < float) : Learning rate. Default: 1e-4 (see utils_lightning.py)
"""
class _CustomMLP(nn.Module):
    def __init__(
        self,
        dims : Union[int, Tuple[int]],
        n_classes : int, 
        optimizer_type : str, 
        batch_size : int ,
        dropout_p : float,
        bn : bool,
        lr : float,
        weight_decay : float,
        l1_strength : float,
        activation_type : str = "relu",
        negative_slope : Optional[float] = None,
        n_hidden_units_list : Optional[List[int]] = None,
        sample_weight : Optional[Union["torch.Tensor", "np.ndarray"]] = None
    ) -> None:
        super(_CustomMLP, self).__init__()
        if isinstance(dims, tuple):
            flattened = 1
            for dim in dims:
                flattened *= dim
            dims = flattened 
        self.dims = dims
        self.n_classes = n_classes
        self.n_hidden_units_list = n_hidden_units_list
        self.optimizer_type = optimizer_type
        self.batch_size = batch_size
        self.dropout_p = dropout_p
        self.bn = bn
        self.lr = lr
        self.weight_decay = weight_decay  
        self.sample_weight = sample_weight
        self.l1_strength = l1_strength
        self.activation_type = activation_type
        self.negative_slope = negative_slope

        for n_hidden_units in self.n_hidden_units_list:
            if n_hidden_units <=0:
                raise ValueError("Number of hidden units can't be less than or equal to 0.")
        
        layers = []

        if self.n_hidden_units_list is None:
            if self.n_classes > 2:
                layers.append(nn.Linear(self.dims, self.n_classes))
            else:
                layers.append(nn.Linear(self.dims, 1))
        
        else:
            layers.append(nn.Linear(self.dims, self.n_hidden_units_list[0]))
            if self.bn:
                layers.append(nn.BatchNorm1d(num_features=self.n_hidden_units_list[0]))
            if activation_type == "tanh":
                layers.append(nn.Tanh())
            elif activation_type == "relu":
                layers.append(nn.ReLU())
            elif activation_type == "leaky_relu":
                if not negative_slope:
                    negative_slope = 1e-2
                layers.append(nn.LeakyReLU(negative_slope=negative_slope))
            else:
                raise ValueError(f"Wrong name {activation_type} for activation function.")
            layers.append(nn.Dropout(self.dropout_p))
            the_rest = self.n_hidden_units_list[1:]
            for idx in range(len(the_rest-1)):
                layers.append(nn.Linear(the_rest[idx], the_rest[idx+1]))
                if self.bn:
                    layers.append(nn.BatchNorm1d(num_features=self.the_rest[idx+1]))
                if activation_type == "tanh":
                    layers.append(nn.Tanh())
                elif activation_type == "relu":
                    layers.append(nn.ReLU())
                elif activation_type == "leaky_relu":
                    if not negative_slope:
                        negative_slope = 1e-2
                    layers.append(nn.LeakyReLU(negative_slope=negative_slope))
                else:
                    raise ValueError(f"Wrong name {activation_type} for activation function.")
                layers.append(nn.Dropout(self.dropout_p))
            
            if self.n_classes > 2:
                layers.append(nn.Linear(the_rest[-1], self.n_classes))
            else:
                layers.append(nn.Linear(the_rest[-1], 1))
        
        if self.n_classes> 2:
            layers.append(nn.Softmax(dim=1))
        else:
            layers.append(nn.Sigmoid())


        self.layers = nn.Sequential(*layers)

    def forward(self, X):
        return self.layers(X)

class CustomMLP(BaseNet):
    def __init__(
        self,
        dims : Union[int, Tuple[int]],
        n_classes : int,
        sample_weight : Optional[Union["torch.Tensor", "np.ndarray"]] = None,
        **hparams : Any
    ) -> None:
        
        entire_keys = [
            'dims', 
            'n_classes', 
            'n_hidden_units_list', 
            'optimizer_type', 
            'batch_size',
            'dropout_p',
            'bn',
            'lr',
            'weight_decay',
            'sample_weight',
            'l1_strength'
            ]
        default_parameter_dict = {
            'n_hidden_units_list' : None
        }
        
        # set default for model-specific hyperparameters
        for argument, default in default_parameter_dict.items():
            if argument not in hparams.keys(): 
                hparams.update({argument : default})
        #set default for model-agnostic hyperparameters
        set_default_parameters(hparams)
        # add hparams : layer_n_dim
        for idx in range(1, len(self.n_hidden_units_list)+1):
            key_str = 'layer_' + str(idx) +'_dim'
            hparams.update({key_str : self.n_hidden_units_list[idx-1]}) 
        # save information of **hparams to self.hparams.
        self.save_hyperparameters(*hparams)
        # n_hidden_units_list
        for n_hidden_units in self.n_hidden_units_list:
            if n_hidden_units <=0:
                raise ValueError(f"Number of hidden units can't be less than or equal to 0.")

        super(CustomMLP, self).__init__(
            dims=dims,
            n_classes=n_classes,
            sample_weight=sample_weight,
            **hparams
        )

        # sanity check. Check if right arguments are declared.
        for key in hparams.keys():
            if key not in entire_keys:
                raise ValueError('Wrong argument is passed to the model.')
        self.model = _CustomMLP(dims=dims, n_classes=n_classes, sample_weight=sample_weight, **hparams)

    def forward(self, X):
        return self.model.forward(X)

########################################################## _CustomMLP & CustomMLP (Ends) ##########################################################

########################################################## _VGG & VGG (Begins) ##########################################################
class _VGG(nn.Module):
    def __init__(
        self,
        dims : Union[Tuple[int], int],
        n_classes : int, 
        cfg : str,
        batch_size : int,
        optimizer_type : str,
        dropout_p : float,
        bn : bool,
        lr : float,
        weight_decay : float,
        l1_strength : float,
        activation_type : str = "relu",
        negative_slope : Optional[float] = None, 
        sample_weight : Optional[Union["torch.Tensor", "np.ndarray"]] = None
    ) -> None: 
        
        super(_VGG, self).__init__()
        if isinstance(dims, int):
            height = width = math.sqrt(dims)
            if (height != int(height)):
                raise ValueError("Wrong dimension in input.")
            dims = (1, height, width)
        if len(dims) == 2:
            height = dims[0]
            width = dims[1]
            dims = (1, height, width)
        if len(dims) == 4:
            dims = dims[1:]
        if len(dims) > 4:
            raise ValueError("dimension of the input can't be higher than 4.")
        self.dims = dims # (in_channels, height, width)
        self.n_classes = n_classes
        self.cfg = cfg
        self.batch_size = batch_size
        self.optimizer_type = optimizer_type
        self.dropout_p = dropout_p
        self.bn = bn
        self.lr = lr
        self.weight_decay = weight_decay
        self.sample_weight = sample_weight
        self.l1_strength = l1_strength
        self.activation_type = activation_type
        self.negative_slope = negative_slope
    
        self.avgpool = nn.AdaptiveAvgPool2d((7, 7))
        layers = []
        layers.append(nn.Linear(512 * 7 * 7, 4096))
        if activation_type == "tanh":
            layers.append(nn.Tanh())
        elif activation_type == "relu":
            layers.append(nn.ReLU())
        elif activation_type == "leaky_relu":
            if not negative_slope:
                negative_slope = 1e-2
            layers.append(nn.LeakyReLU(negative_slope=negative_slope))
        else:
            raise ValueError(f"Wrong name {activation_type} for activation function.")

        layers.append(nn.Dropout(p=self.dropout_p))
        layers.append(nn.Linear(4096, 4096))
        if activation_type == "tanh":
            layers.append(nn.Tanh())
        elif activation_type == "relu":
            layers.append(nn.ReLU())
        elif activation_type == "leaky_relu":
            if not negative_slope:
                negative_slope = 1e-2
            layers.append(nn.LeakyReLU(negative_slope=negative_slope))
        else:
            raise ValueError(f"Wrong name {activation_type} for activation function.")
        layers.append(nn.Dropout(p=self.dropout_p))
        layers.append(nn.Linear(4096, n_classes))
        
        self.classifier = nn.Sequential(*layers)

        self.cfgs: Dict[str, List[Union[str, int]]] = {
        "NoMaxPool" :[64,  128, 256, 256,  512, 512,  512, 512], #only for debugging.
        "VGG11": [64, "M", 128, "M", 256, 256, "M", 512, 512, "M", 512, 512, "M"],
        "VGG13": [64, 64, "M", 128, 128, "M", 256, 256, "M", 512, 512, "M", 512, 512, "M"],
        "VGG16": [64, 64, "M", 128, 128, "M", 256, 256, 256, "M", 512, 512, 512, "M", 512, 512, 512, "M"],
        "VGG19": [64, 64, "M", 128, 128, "M", 256, 256, 256, 256, "M", 512, 512, 512, 512, "M", 512, 512, 512, 512, "M"],
        }
        
    
    def make_layers(
        self,
        cfg : List[Union[str,int]],
        bn : bool=True
    ) -> nn.Sequential:
        layers = []
        in_channels = (self.dims)[0]
        for v in cfg:
            if v=="M":
                layers.append(nn.MaxPool2d(kernel_size=2, stride=2))
            else:
                v = cast(int, v)
                layers.append(nn.Conv2d(in_channels, v, kernel_size=3, padding=1))
                if bn:
                    layers.append(nn.BatchNorm2d(v))
                    
                if self.activation_type == "tanh":
                    layers.append(nn.Tanh())
                elif self.activation_type == "relu":
                    layers.append(nn.ReLU())
                elif self.activation_type == "leaky_relu":
                    if not negative_slope:
                        negative_slope = 1e-2
                    layers.append(nn.LeakyReLU(negative_slope=self.negative_slope))
                else:
                    raise ValueError(f"Wrong name {self.activation_type} for activation function.")
                in_channels = v

        return nn.Sequential(*layers)    


    def forward(self, X) -> torch.Tensor:
        out = self.make_layers(self.cfgs[self.cfg], self.bn)(X)
        out = self.avgpool(out)
        out = torch.flatten(out, 1)
        out = self.classifier(out) 
        return out

class VGG(ConvNet):

    def __init__(
        self,
        dims : Union[int, Tuple[int]],
        n_classes : int,
        sample_weight : Optional[Union["torch.Tensor", "np.ndarray"]] = None,
        cfg : str ="VGG11",
        **hparams
    ) -> None:
        

        entire_keys = [ 
            'dims',
            'n_classes', 
            'cfg', 
            'optimizer_type', 
            'batch_size',
            'dropout_p',
            'bn',
            'lr',
            'weight_decay',
            'sample_weight',
            'l1_strength',
            'activation_type',
            'negative_slope'
            ]

        default_parameter_dict = {}
    
        # set default for model-specific hyperparameters
        for argument, default in default_parameter_dict.items():
            if argument not in hparams.keys(): 
                hparams.update({argument : default})
        #set default for model-agnostic hyperparameters
        set_default_parameters(hparams)
        # save information of **hparams to self.hparams.
        self.save_hyperparameters(*hparams)

        super(VGG, self).__init__(
            dims=dims,
            n_classes=n_classes,
            sample_weight=sample_weight,
            model_type="VGG",
            cfg=cfg,
            **hparams
        )
        # sanity check. Check if right arguments are declared.
        for key in hparams.keys():
            if key not in entire_keys:
                raise ValueError('Wrong argument is passed to the model.')
        self.model = _VGG(
            dims=dims,
            n_classes=n_classes,
            sample_weight=sample_weight,
            cfg=cfg,
            **hparams
        )

    def forward(self, X):
        return self.model.forward(X)

    def help(self) -> None :
        whoami = "VGG Network\n\n"
        helptext = "Parameters ------------------------------------------------------------------------------------------------------- \n"
        helptext += "n_classes: int \t Number of classes in the classification task. If it is binary classification, n_classes=2\n"
        helptext += "cfg: str\t The configuration of VGG. 'VGG11', 'VGG13', 'VGG16', and 'VGG19' are available.\n"
        helptext += "In defining hyperparameters and optimizer_type, we recommend you to define a dictionary.\n"
        helptext += "Define a dictionary containing :\n\t-'optimizer_type': str (default='adam')\n\t-'dropout_p' : 0 <= float < 1(default : 0)\n\t-'weight_decay' : 0<= float < 1(default=0)\n\t-'lr' : float(default=1e-4)."
        
        text= whoami + helptext
        print(text)
        
class _UNet(nn.Module):
    """
    Paper: `U-Net: Convolutional Networks for Biomedical Image Segmentation
    <https://arxiv.org/abs/1505.04597>`
    Args:
        n_classes: Number of output classes required
        dims : Dimension of the input
        num_layers: Number of layers in each side of U-net (default 5)
        features_start: Number of features in first layer (default 64)
        bilinear: Whether to use bilinear interpolation or transposed convolutions (default) for upsampling.
    """

    def __init__(
        self,
        n_classes : int, 
        dims : Union[int, Tuple[int]],
        batch_size : int,
        optimizer_type : str,
        lr : float,
        weight_decay : float,
        n_layers : int, 
        bn : bool,
        features_start : int, 
        bilinear : bool,
        l1_strength : float,
        activation_type : str = "relu",
        negative_slope : Optional[float] = None,
        model_type : str = "UNet", # don't change this ,
        sample_weight : Optional[Union["torch.Tensor", "np.ndarray"]] = None
    ):
        self.n_classes = n_classes
        self.batch_size = batch_size
        self.optimizer_type = optimizer_type
        self.bn = bn
        self.lr = lr
        self.weight_decay = weight_decay
        self.l1_strength = l1_strength
        self.n_layers = n_layers
        self.features_start = features_start
        self.bilinear = bilinear
        self.model_type = model_type
        self.sample_weight = sample_weight
        self.activation_type = activation_type 
        self.negative_slope = negative_slope

        if self.n_layers < 1:
            raise ValueError(f"n_layers = {self.n_layers}, expected: n_layers > 0")

        if isinstance(dims, int):
            height = width = math.sqrt(dims)
            if (height != int(height)):
                raise ValueError("Wrong dimension in input.")
            dims = (1, height, width)
        if len(dims) == 2:
            height = dims[0]
            width = dims[1]
            dims = (1, height, width)
        if len(dims) == 4:
            dims = dims[1:]
        if len(dims) > 4:
            raise ValueError("dimension of the input can't be higher than 4.")
        self.dims = dims # (in_channels, height, width)
        self.input_channels = self.dims[0]
        super().__init__()

        layers = [DoubleConv(self.input_channels, self.features_start, 
                             bn=self.bn, activation_type=activation_type, negative_slope=negative_slope)]

        feats = self.features_start
        for _ in range(self.n_layers - 1):
            layers.append(Down(feats, feats * 2))
            feats *= 2

        for _ in range(self.n_layers - 1):
            layers.append(Up(feats, feats // 2, self.bilinear))
            feats //= 2

        layers.append(nn.Conv2d(feats, n_classes, kernel_size=1))

        self.layers = nn.ModuleList(layers)

    def forward(self, x):
        xi = [self.layers[0](x)]
        # Down path
        for layer in self.layers[1 : self.n_layers]:
            xi.append(layer(xi[-1]))
        # Up path
        for i, layer in enumerate(self.layers[self.n_layers : -1]):
            xi[-1] = layer(xi[-1], xi[-2 - i])
        return self.layers[-1](xi[-1])


class UNet(ConvNet):
    def __init__(
        self,
        dims : Union[int, Tuple[int]],
        n_classes: int,
        model_type : str = "UNet", # don't change this
        sample_weight : Optional[Union["torch.Tensor", "np.ndarray"]] = None,
        **hparams : Any
    ):

        entire_keys = [
            'n_classes', 
            'model_type',
            'dims', 
            'n_layers', 
            'bn',
            'features_start', 
            'bilinear',
            'batch_size' ,
            'optimizer_type' ,
            'lr' ,
            'weight_decay',
            'sample_weight',
            'l1_strength',
            'activation_type',
            'negative_slope'
        ]

        default_parameter_dict = {
            'n_layers' : 3,
            'features_start' : 64,
            'bilinear' : False
        }
         
        # set default for model-specific hyperparameters
        for argument, default in default_parameter_dict.items():
            if argument not in hparams.keys(): 
                hparams.update({argument : default})
        #set default for model-agnostic hyperparameters
        set_default_parameters(hparams)
        # for some reason, we don't use dropout here
        hparams.pop('dropout_p')
        # save information of **hparams to self.hparams.
        self.save_hyperparameters(*hparams)

        super(UNet, self).__init__(
            n_classes=n_classes, 
            dims=dims, 
            sample_weight=sample_weight,
            model_type="UNet",
            **hparams
        )
        # sanity check. Check if right arguments are declared.
        for key in hparams.keys():
            if key not in entire_keys:
                raise ValueError(f'Wrong argument is passed to the model.')
        self.model = _UNet(
            n_classes=n_classes, 
            dims=dims, 
            sample_weight=sample_weight,
            **hparams
        )

    def forward(self, X):
        return self.model.forward(X)
    
    
"""

TODO : Implement GoogleNet
TODO : Implement ResNet
TODO : Implement YOLO

"""


import torch
import torch.nn as nn
import torch.nn.functional as F
import pytorch_lightning as pl 
from typing import Union, List, Dict, Any, Optional, cast

class DoubleConv(nn.Module):
    """[ Conv2d => BatchNorm (optional) => ReLU ] x 2."""

    def __init__(
        self, 
        in_ch: int, 
        out_ch: int,
        bn : bool = True,
        activation_type : str = "relu",
        negative_slope : Optional[float] = None 
    ):
        super().__init__()
        
        layers = []
        layers.append(nn.Conv2d(in_ch, out_ch, kernel_size=3, padding=1))
        if bn:
            layers.append(nn.BatchNorm2d(out_ch))
        if activation_type == "tanh":
            layers.append(nn.Tanh())
        elif activation_type == "relu":
            layers.append(nn.ReLU(inplace=True))
        elif activation_type == "leaky_relu":
            if not negative_slope:
                negative_slope = 1e-2
            layers.append(nn.LeakyReLU(negative_slope=negative_slope, inplace=True))
        else:
            raise ValueError(f"Wrong name {activation_type} for activation function.")
        
        layers.append(nn.Conv2d(out_ch, out_ch, kernel_size=3, padding=1))
        if bn:
            layers.append(nn.BatchNorm2d(out_ch))
        if activation_type == "tanh":
            layers.append(nn.Tanh())
        elif activation_type == "relu":
            layers.append(nn.ReLU(inplace=True))
        elif activation_type == "leaky_relu":
            if not negative_slope:
                negative_slope = 1e-2
            layers.append(nn.LeakyReLU(negative_slope=negative_slope, inplace=True))
        else:
            raise ValueError(f"Wrong name {activation_type} for activation function.")
        
        
        self.net = nn.Sequential(*layers)

    def forward(self, x):
        return self.net(x)


class Down(nn.Module):
    """Downscale with MaxPool => DoubleConvolution block."""

    def __init__(self, in_ch: int, out_ch: int):
        super().__init__()
        self.net = nn.Sequential(nn.MaxPool2d(kernel_size=2, stride=2), DoubleConv(in_ch, out_ch))

    def forward(self, x):
        return self.net(x)


class Up(nn.Module):
    """Upsampling (by either bilinear interpolation or transpose convolutions) followed by concatenation of feature
    map from contracting path, followed by DoubleConv."""

    def __init__(
        self, 
        in_ch: int, 
        out_ch: int, 
        bilinear: bool
    ):
        super().__init__()
        self.upsample = None
        if bilinear:
            self.upsample = nn.Sequential(
                nn.Upsample(scale_factor=2, mode="bilinear", align_corners=True),
                nn.Conv2d(in_ch, in_ch // 2, kernel_size=1),
            )
        else:
            self.upsample = nn.ConvTranspose2d(in_ch, in_ch // 2, kernel_size=2, stride=2)

        self.conv = DoubleConv(in_ch, out_ch)

    def forward(self, x1, x2):
        x1 = self.upsample(x1)

        # Pad x1 to the size of x2
        diff_h = x2.shape[2] - x1.shape[2]
        diff_w = x2.shape[3] - x1.shape[3]

        x1 = F.pad(x1, [diff_w // 2, diff_w - diff_w // 2, diff_h // 2, diff_h - diff_h // 2])

        # Concatenate along the channels axis
        x = torch.cat([x2, x1], dim=1)
        return self.conv(x)




import torch
import torch.nn as nn
import pytorch_lightning as pl
from astrape.utilities.utils import *
from astrape.exceptions.exceptions import *
from astrape.constants.astrape_constants import *
from typing import Union, List, Dict, Any, Optional, cast, Tuple
import math 
import numpy as np

applicable_val_metric_list = APPLICABLE_VAL_METRIC_LIST
applicable_optimizer_type_list = APPLICABLE_OPTIMIZER_TYPE_LIST


class BaseNet(pl.LightningModule):
    r"""BaseNet. Use this for models that don't deal with image data. 

        Attributes:
            
            dims (int or tuple of int): Dimension of data with # of samples removed.
            
            n_classes (int or None): Number of classes in classification task. None if it is a regression task.
            
            config (dict): Hyperparameter configurations.

            sample_weight (torch.Tensor or np.ndarray): Sample weights.

            **kwargs (Any): Attributes of LightningModule

        Note:

            BaseNet is the parent class for MLPs(MLP, ContractingMLP, CustomMLP). It is also
            compatible with other MLPs that are written in pytorch_lightning.

            When inheriting BaseNet, you only need to specify the structure of your MLP because 
            other magics are defined in BaseNet.  

            BaseNet can perform both classification and regression tasks. If you specify n_classes, 
            BaseNet will recognize the task as a classification problem. If n_classes is not passed
            as a parameter for BaseNet, it will understand that the task is a regression problem.
    """

    def __init__(
        self,
        dims : Union[int, Tuple[int]],
        n_classes : Optional[int],
        l1_strength : float = 0.0,
        sample_weight : Optional[Union["np.ndarray", "torch.Tensor"]] = None,
        **config : Any
    ) -> None:
        r"""
         Args: 
            
            dims (int or tuple of int): The dimension of the data. e.g., 4, (8,8,1)

            n_classes (int, optional): Number of classes (labels) in the classification task. Do not specify this argument (set is as None) when you are performing regression tasks.

            **config (Any) : Hyperparameters that can be shared among different models. e.g., optimizer_type, lr, weight_decay, batch_size, bn, dropout_p, etc.
        """
        super(BaseNet, self).__init__()
        if isinstance(dims, tuple):
            flattened = 1
            for dim in dims:
                flattened *= dim
            dims = flattened 

        self.dims = dims
        self.n_classes = n_classes if n_classes is not None else None 
        self.config = config
        self.sample_weight = sample_weight

        config.update({'l1_strength' : l1_strength})
            
        self.save_hyperparameters(*config)
        

    @property
    def criterion(self):
        r"""Criterion for loss function.

        Returns:

            See astrape.utilities.utils for details.
        """
        if self.n_classes is None:
            if 'regression_metric' in self.config.keys():
                if self.config['regression_metric'] == "mse":
                    criterion = MSELoss()
                elif self.config['regression_metric'] == "r2":
                    criterion = R2Score()
                elif self.config['regression_metric'] == "rmse":
                    criterion = RMSELoss()
                else:
                    raise NotImplementedError("Only MSE loss, R2 score, and RMSE are supported in expytorch_lightning v 0.0.0.")
            else:
                criterion = MSELoss()
        else:
            criterion = CELossMetrics(self.n_classes) if self.n_classes > 2 else BCELossMetrics()
        return criterion

    def training_step(self, train_batch, batch_idx):

        x_batch, y_batch = train_batch
        Yhat = self.forward(x_batch)
        y_batch = self.sample_weight * y_batch
        

        if self.n_classes is None: # regression
            loss = self.criterion(Yhat.squeeze(), y_batch)
            # L1 regularizer
            if self.hparams.l1_strength > 0:
                l1_reg = self.linear.weight.abs().sum()
                loss += self.hparams.l1_strength * l1_reg
            self.log('train/loss', loss)
            return {'loss' : loss}
        else: # classification
            loss, acc, auc = self.criterion(Yhat.squeeze(), y_batch.long())

            self.log('train/loss', loss)
            self.log('train/acc', acc)
            self.log('train/auc', auc)
            return {'loss' : loss, 'acc' : acc, 'auc' : auc}
    
    def validation_step(self, val_batch, batch_idx):
        x_batch, y_batch = val_batch
        
        Yhat = self.forward(x_batch)
        if self.n_classes is None:
            loss = self.criterion(Yhat.squeeze(), y_batch)
            # L1 regularizer
            if self.hparams.l1_strength > 0:
                l1_reg = self.linear.weight.abs().sum()
                loss += self.hparams.l1_strength * l1_reg

            
            self.log('val/loss', loss)
            return {'loss' : loss}
        else:    
            loss, acc, auc = self.criterion(Yhat.squeeze().type_as(y_batch), y_batch)

            self.log('val/loss', loss)
            self.log('val/acc', acc)
            self.log('val/auc', auc)
            return {'loss' : loss, 'acc' : acc, 'auc' : auc}
    
    def predict_step(self, pred_batch, batch_idx, dataloader_idx=0):
        x_batch, y_batch = pred_batch
        
        Yhat = self.forward(x_batch)
        if self.n_classes is None:
            loss = self.criterion(Yhat.squeeze(), y_batch)
            # L1 regularizer
            if self.hparams.l1_strength > 0:
                l1_reg = self.linear.weight.abs().sum()
                loss += self.hparams.l1_strength * l1_reg

            
            self.log('predict/loss', loss)
            return {'loss' : loss}
        else:    
            loss, acc, auc = self.criterion(Yhat.squeeze(), y_batch.type(torch.LongTensor))

            self.log('predict/loss', loss)
            self.log('predict/acc', acc)
            self.log('predict/auc', auc)
            return {'loss' : loss, 'acc' : acc, 'auc' : auc}

    def test_step(self, test_batch, batch_idx, data_loader_idx=0):
        x_batch, y_batch = test_batch
        
        Yhat = self.forward(x_batch)
        if self.n_classes is None:
            loss = self.criterion(Yhat.squeeze(), y_batch)
            # L1 regularizer
            if self.hparams.l1_strength > 0:
                l1_reg = self.linear.weight.abs().sum()
                loss += self.hparams.l1_strength * l1_reg

            
            self.log('test/loss', loss)
            return {'loss' : loss}
        else:    
            loss, acc, auc = self.criterion(Yhat.squeeze(), y_batch.type(torch.LongTensor))

            self.log('test/loss', loss)
            self.log('test/acc', acc)
            self.log('test/auc', auc)
            return {'loss' : loss, 'acc' : acc, 'auc' : auc}

    def on_after_batch_transfer(self, batch, dataloader_idx=0):
        # modify this part if you want to perform model reweigthing.
        self.sample_weight = torch.ones_like(batch[1]).type_as(batch[1])
        return batch
        
    def reset_weights(self, module):
        if isinstance(module, nn.Linear):
            module.reset_parameters()
            
    def _init_weights(self, module):
        if isinstance(module, nn.Linear):
            nn.init.kaiming_normal_(module.weight)
        elif isinstance(module, nn.BatchNorm1d):
            nn.init.constant_(module.weight, 1)
            nn.init.constant_(module.bias, 0)

    def configure_optimizers(self):

        #optimizer_type_check(applicable_optimizer_type_list, self.config.optimizer_type)
        if self.hparams.optimizer_type == "adam":
            optimizer = torch.optim.Adam(self.parameters(),
                                     lr=self.hparams.lr,
                                     weight_decay=self.hparams.weight_decay)
        elif self.hparams.optimizer_type == "sgd":
            optimizer = torch.optim.SGD(self.parameters(),
                                        lr=self.hparams.lr)
        
        return optimizer


class ConvNet(pl.LightningModule):
    r"""ConvNet. Use this when using CNNs. 

    Note:

        ConvNet is the parent class for CNNs(VGG, UNet). It is also compatible with other CNNs
        that are written in pytorch_lightning. 

        When inheriting ConvNet, you only need to specify the structure of your MLP because 
        other magics are defined in ConvNet. 

        Currently in expytorch_lightning v. 0.0.0, ConvNet only supports basic classification
        tasks such as classification using CE loss or dice loss. Other operations will be 
        updated soon(Be tuned!).

    Attributes:

        model_type (LightningModule): The class of the model that you will use.

        n_classes (int, optional): Number of classes in the classification task. Leave it as None in regression tasks.

        sample_weight (torch.Tensor or np.ndarray): Sample weights.

        config (dict) : Hyperparameters that can be shared among different models. e.g., optimizer_type, lr, weight_decay, batch_size, bn, dropout_p, etc.
"""
    def __init__(
        self,
        model_type : "pl.LightningModule",
        n_classes : Optional[int] = None,
        sample_weight : Optional[Union["np.ndarray", "torch.Tensor"]] = None,
        **config : Any
    )->None:
        r"""
        Args: 

            model_type (LightningModule): The class of the model that you will use.

            n_classes (int, optional): Number of classes in the classification task. Leave it as None in regression tasks.
                Default: ``None``
            
            sample_weight (torch.Tensor or np.ndarray): Sample weights.
                Default: ``None``

            **config (Any) : Hyperparameters that can be shared among different models. e.g., optimizer_type, lr, weight_decay, batch_size, bn, dropout_p, etc.
        """
        super(ConvNet, self).__init__()

        if 'optimizer_type' not in config.keys():
            config.update({'optimizer_type' : 'adam'})
        
        self.model_type = model_type
        self.n_classes = n_classes
        self.sample_weight = sample_weight
        self.save_hyperparameters(*config)
        

    @property
    def criterion(self):   
        r"""Criterion for loss function.

        Returns:

            See astrape.utilities.utils for details.
        """
        if self.n_classes is None:
            if 'regression_metric' in self.config.keys():
                if self.config['regression_metric'] == "mse":
                    criterion = MSELoss()
                elif self.config['regression_metric'] == "r2":
                    criterion = R2Score()
                elif self.config['regression_metric'] == "rmse":
                    criterion = RMSELoss()
                else:
                    raise NotImplementedError("Only MSE loss, R2 score, and RMSE are supported in expytorch_lightning v 0.0.0.")
            else:
                criterion = MSELoss()
        else:
            if self.model_type == "UNet":
                criterion = CEDiceLossMetrics(self.n_classes)
            else:
                criterion = CELossMetrics(self.n_classes) if self.n_classes > 2 else BCELossMetrics()
        return criterion

    @staticmethod
    def reshape_x(x_batch):
        dims = x_batch.shape[1:]
        
        if len(dims) == 1:
            width = height = math.sqrt(dims[0])
            if (height != int(height)):
                raise ValueError("Wrong dimension for the input.")
            x_batch = x_batch.reshape(-1, 1, int(height), int(width))
        elif len(dims) == 2:
            height = dims[0]
            width = dims[1]
            x_batch = x_batch.reshape(-1, 1, int(height), int(width))
        elif len(dims) == 3:
            x_batch = x_batch.reshape(-1, int(dims[0]), int(dims[1]), int(dims[2]))
        else:
            raise ValueError("Wrong dimension for the input.")

        return x_batch

    def on_before_batch_transfer(self, batch: Any, dataloader_idx=0):
        batch[0] = self.reshape_x(batch[0])
        return batch
    
    def on_after_batch_transfer(self, batch, dataloader_idx=0):
        # modify this part if you want to perform model reweigthing.
        self.sample_weight = torch.ones_like(batch[1]).type_as(batch[1])
        return batch
    
    def training_step(self, train_batch, batch_idx):
        x_batch, y_batch = train_batch
        x_batch = self.reshape_x(x_batch)
            
        y_batch = self.sample_weight * y_batch

        Yhat = self.forward(x_batch)
        if self.n_classes is None:
            loss = self.criterion(Yhat.squeeze(), y_batch)
            self.log('train/loss', loss)
            return {'loss' : loss}
        else:    
            loss, acc, auc = self.criterion(Yhat.squeeze(), y_batch.type(torch.LongTensor))

            self.log('train/loss', loss)
            self.log('train/acc', acc)
            self.log('train/auc', auc)
            return {'loss' : loss, 'acc' : acc, 'auc' : auc}

        
    def validation_step(self, val_batch, batch_idx):
        x_batch, y_batch = val_batch
        x_batch = self.reshape_x(x_batch)
        Yhat = self.forward(x_batch)
        if self.n_classes is None:
            loss = self.criterion(Yhat.squeeze(), y_batch)
            self.log('val/loss', loss)
            return {'loss' : loss}
        else:    
            loss, acc, auc = self.criterion(Yhat.squeeze(), y_batch.type(torch.LongTensor))

            self.log('val/loss', loss)
            self.log('val/acc', acc)
            self.log('val/auc', auc)
            return {'loss' : loss, 'acc' : acc, 'auc' : auc}

    def predict_step(self, pred_batch, batch_idx, dataloader_idx=0):
        x_batch, y_batch = pred_batch
        x_batch = self.reshape_x(x_batch)
        Yhat = self.forward(x_batch)
        if self.n_classes is None:
            loss = self.criterion(Yhat.squeeze(), y_batch)
            self.log('pred/loss', loss)
            return {'loss' : loss}
        else:    
            loss, acc, auc = self.criterion(Yhat.squeeze(), y_batch.type(torch.LongTensor))

            self.log('pred/loss', loss)
            self.log('pred/acc', acc)
            self.log('pred/auc', auc)
            return {'loss' : loss, 'acc' : acc, 'auc' : auc}

    def test_step(self, test_batch, batch_idx):
        x_batch, y_batch = test_batch
        x_batch = self.reshape_x(x_batch)
        Yhat = self.forward(x_batch)
        if self.n_classes is None:
            loss = self.criterion(Yhat.squeeze(), y_batch)
            self.log('test/loss', loss)
            return {'loss' : loss}
        else:    
            loss, acc, auc = self.criterion(Yhat.squeeze(), y_batch.type(torch.LongTensor))

            self.log('test/loss', loss)
            self.log('test/acc', acc)
            self.log('test/auc', auc)
            return {'loss' : loss, 'acc' : acc, 'auc' : auc}

    def reset_weights(self, module):
        if isinstance(module, nn.Linear):
            module.reset_parameters()
            
    def _init_weights(self, module) -> None:
        if isinstance(module, nn.Linear):
            nn.init.kaiming_normal_(module.weight)
        elif isinstance(module, nn.BatchNorm1d):
            nn.init.constant_(module.weight, 1)
            nn.init.constant_(module.bias, 0)
        elif isinstance(module, nn.Conv2d):
            nn.init.kaiming_normal_(module.weight, mode="fan_out", nonlinearity="relu")
            if module.bias is not None:
                nn.init.constant_(module.bias, 0)

    def configure_optimizers(self):
        if self.hparams.optimizer_type == "adam":
            optimizer = torch.optim.Adam(self.parameters(),
                                     lr=self.hparams.lr,
                                     weight_decay=self.hparams.weight_decay)
        elif self.hparams.optimizer_type == "sgd":
            optimizer = torch.optim.SGD(self.parameters(),
                                        lr=self.hparams.lr)

        return optimizer


import sys
import warnings

import pkg_resources
from pkg_resources import parse_version


from .experiment import Experiment
from .model_selection import CrossValidation

MIN_TORCH_VERSION = '1.1.0'
MIN_LIGHTNING_VERSION = '1.6.0'

try:
    import torch
except ModuleNotFoundError:
    raise ModuleNotFoundError(
        "No module named 'torch', and astrape depends on PyTorch "
        "(aka 'torch'). "
        "Visit https://pytorch.org/ for installation instructions.")


try:
    import pytorch_lightning
except ModuleNotFoundError:
    raise ModuleNotFoundError(
        "No module named 'pytorch-lightning', and astrape depends on PyTorch-Lightning "
        "(aka 'pytorch_lightning'). "
        "Visit https://www.pytorchlightning.ai/ for installation instructions.")

torch_version = pkg_resources.get_distribution('torch').version
if parse_version(torch_version) < parse_version(MIN_TORCH_VERSION):
    msg = ('astrape depends on a newer version of PyTorch (at least {req}, not '
           '{installed}). Visit https://pytorch.org for installation details')
    raise ImportWarning(msg.format(req=MIN_TORCH_VERSION, installed=torch_version))

lightning_version = pkg_resources.get_distribution('pytorch_lightning').version
if parse_version(lightning_version) < parse_version(MIN_LIGHTNING_VERSION):
    msg = ('astrape depends on a newer version of PyTorch (at least {req}, not '
           '{installed}). Visit https://www.pytorchlightning.ai/ for installation details')
    raise ImportWarning(msg.format(req=MIN_LIGHTNING_VERSION, installed=lightning_version))


from astrape.experiment import Experiment
from astrape.model_selection import CrossValidation
from astrape.project import Project
from astrape.utilities.dataloader_lightning import DataModule
from astrape.utilities.utils_lightning import initialize_datamodule, initialize_model
from astrape.models.models_lightning import MLP, ContractingMLP, CustomMLP, VGG, UNet
from astrape.base.experiment_base import BaseExperiment
from astrape.constants.astrape_constants import *
__all__ = [
    'Experiment',
    'CrossValidation',
    'Project',
    'DataModule',
    'MLP',
    'ContractingMLP',
    'CustomMLP',
    'VGG',
    'UNet',
    'BaseExperiment'
]

__import__("pkg_resources").declare_namespace(__name__)

try:
    __version__ = pkg_resources.get_distribution('astrape').version
except:
    __version__ = 'n/a'


from ast import Assert
from math import log, sqrt
import torch
import numpy as np
import torch.nn as nn
import torch.nn.functional as F
from torch import Tensor
import matplotlib.pyplot as plt
from torchmetrics import AUROC
from torchmetrics.functional import dice_score
from typing import Union, List, Dict, Any, Optional

def column_or_1d(y):
    y = torch.as_tensor(y)
    shape = y.shape
    if (len(shape) == 1) or (len(shape) == 2 and shape[1] == 1):
        return torch.ravel(y)
    
    raise ValueError(f"y should be a 1d tensor, got a tensor of shape {shape} instead.")

class BCELossMetrics():
    def __init__(self):
        self.loss_function = nn.BCELoss()
    
    @staticmethod
    def accuracy(y_hat, labels):
        with torch.no_grad():
            y_tilde = (y_hat > 0.5).int()
            accuracy = (y_tilde == labels.int()).float().mean().item()
        return accuracy

    @staticmethod
    def auc(y_hat, labels):
        with torch.no_grad():
            auc_fn = AUROC()
            auc = auc_fn(y_hat, labels.int()).float()
        return auc

    def __call__(
        self, 
        y_hat : "torch.Tensor", 
        labels : "torch.Tensor"
    ):
        
        loss = self.loss_function(y_hat, labels.reshape(-1).long()).squeeze()
        accuracy = self.accuracy(y_hat, labels)
        auc = self.auc(y_hat, labels.int())
        return loss, accuracy, auc

class CELossMetrics():
    def __init__(self, n_classes):
        self.loss_function = nn.CrossEntropyLoss()
        self.n_classes = n_classes
    @staticmethod
    def accuracy(y_hat, labels):
        with torch.no_grad():
            y_tilde = y_hat.argmax(axis=1)
            accuracy = (y_tilde == labels).float().mean().item()
        return accuracy

    
    def auc(self,y_hat, labels):
        with torch.no_grad():
            auc_fn = AUROC(num_classes=self.n_classes)
            auc = auc_fn(y_hat, labels.int()).float()
        return auc
    
    def __call__(
        self, 
        y_hat, 
        labels : "torch.Tensor"
    ):
        
        loss = self.loss_function(y_hat, labels.reshape(-1).long()).squeeze()
        accuracy = self.accuracy(y_hat, labels)
        auc = self.auc(y_hat, labels)
        return loss, accuracy, auc

class MSELoss():
    def __init__(self):
        pass
    def __call__(
        self,
        y_hat : "torch.Tensor", 
        labels : "torch.Tensor"
    ):
        criterion = nn.MSELoss()
        loss = criterion(y_hat, labels.reshape(-1)).squeeze()
        return loss
class RMSELoss():
    def __init__(self):
        pass

    def __call__(
        self,
        y_hat : "torch.Tensor", 
        labels : "torch.Tensor"
    ):        
        criterion = nn.MSELoss()
        loss = torch.sqrt(criterion(y_hat, labels.reshape(-1))).squeeze()
        return loss

class R2Score():
    def __init__(self):
        pass
    def __call__(
        self, 
        y_hat, 
        labels : "torch.Tensor", 
        sample_weight : Optional["torch.Tensor"] = None,
        multioutput : Union[str, "torch.Tensor"] = "uniform_average"
    ):
        if len(y_hat) < 2:
            raise ValueError("R^2 score is not well-defined with less than 2 samples.")
        if sample_weight is not None:
            sample_weight = column_or_1d(sample_weight).type_as(labels)
            weight = torch.unsqueeze(sample_weight, dim=1).type_as(labels)
        else:
            weight = torch.ones_like(labels).unsqueeze(dim=1).type_as(labels)

        numerator = torch.sum(weight * (labels - y_hat)**2, dim=0).float()
        weighted_sum = (weight * labels.unsqueeze(dim=1)).squeeze()
        denominator = weight * torch.sum(labels - torch.mean(weighted_sum, dim=0)**2, dim=0).float()
        
        nonzero_denominator = denominator != 0
        nonzero_numerator = numerator != 0
        valid_score = nonzero_denominator & nonzero_numerator
        output_scores = torch.ones(size=labels.shape[1])
        output_scores[valid_score] = 1 - (numerator[valid_score]/denominator[valid_score])
        # arbitrary set to zero to avoid -inf scores, having a constant
        # labels are not interesting for scoring a regression anyway
        output_scores[nonzero_numerator & ~nonzero_denominator] = 0.0
        if isinstance(multioutput, str):
            if multioutput == "raw_values":
                # return scores individually
                return output_scores
            elif multioutput == "uniform_average":
                # passing None as weights results is uniform mean
                avg_weights = None
            elif multioutput == "variance_weighted":
                avg_weights = denominator.type_as(weight)
                # avoid fail on constant y or one-element arrays
                if not np.any(nonzero_denominator):
                    if not np.any(nonzero_numerator):
                        return torch.ones_like(labels).type_as(labels)
                    else:
                        return torch.zeros_like(labels).type_as(labels)
        else:
            avg_weights = multioutput

        output_weighted_sum = avg_weights * output_scores
        loss = torch.mean(output_weighted_sum)
        return loss


class CEDiceLossMetrics():
    def __init__(self, n_classes):
        self.loss_function = nn.CrossEntropyLoss()
        self.n_classes = n_classes
    @staticmethod
    def accuracy(y_hat, labels):
        with torch.no_grad():
            y_tilde = y_hat.argmax(axis=1)
            accuracy = (y_tilde == labels).float().mean().item()
        return accuracy

    
    def auc(self,y_hat, labels):
        with torch.no_grad():
            auc_fn = AUROC(num_classes=self.n_classes)
            auc = auc_fn(y_hat, labels.int()).float()
        return auc

    def __call__(
        self, 
        y_hat, 
        labels : "torch.Tensor"
    ):
        
        loss = self.loss_function(y_hat, labels.reshape(-1)).squeeze()
        """
        loss += dice_loss(
                    F.softmax(y_hat, dim=1).float(),
                    F.one_hot(labels, self.n_classes).permute(0,3,1,2).float(),
                    multiclass=True
                )
        
        """
        loss += dice_score(F.softmax(y_hat, dim=1).float(), labels.reshape(-1)).squeeze()
        accuracy = self.accuracy(y_hat, labels)
        auc = self.auc(y_hat, labels)
        return loss, accuracy, auc



def dice_coeff(
    input: Tensor,
    target: Tensor,
    reduce_batch_first: bool = False, epsilon=1e-6
):
    # Average of Dice coefficient for all batches, or for a single mask
    if (input.size() == target.size()):
        raise AssertionError("Size mismatch between input and target.")
    if (input.dim() == 2 and reduce_batch_first):
        raise ValueError(f'Dice: asked to reduce batch but got tensor without batch dimension (shape {input.shape})')

    if input.dim() == 2 or reduce_batch_first:
        inter = torch.dot(input.reshape(-1), target.reshape(-1))
        sets_sum = torch.sum(input) + torch.sum(target)

        if sets_sum.item() == 0:
            sets_sum = 2 * inter

        return (2 * inter + epsilon) / (sets_sum + epsilon)

    else:
        # compute and average metric for each batch element
        dice = 0
        for i in range(input.shape[0]):
            dice += dice_coeff(input[i, ...], target[i, ...])

        return dice / input.shape[0]


def multiclass_dice_coeff(
    input: Tensor, 
    target: Tensor, 
    reduce_batch_first: bool = False, 
    epsilon=1e-6
):
    # Average of Dice coefficient for all classes
    if (input.size() == target.size()):
        raise AssertionError("Size mismatch between input and target.")
    dice = 0
    for channel in range(input.shape[1]):
        dice += dice_coeff(input[:, channel, ...], target[:, channel, ...], reduce_batch_first, epsilon)

    return dice / input.shape[1]


def dice_loss(
    input: Tensor, 
    target: Tensor, 
    multiclass: bool = False
):
    # Dice loss (objective to minimize) between 0 and 1
    if (input.size() == target.size()):
        raise AssertionError("Size mismatch between input and target.")
    fn = multiclass_dice_coeff if multiclass else dice_coeff

    return 1 - fn(input, target, reduce_batch_first=True)
    


def plot_img_and_mask(img, mask):
    classes = mask.shape[0] if len(mask.shape) > 2 else 1
    fig, ax = plt.subplots(1, classes + 1)
    ax[0].set_title('Input image')
    ax[0].imshow(img)
    if classes > 1:
        for i in range(classes):
            ax[i + 1].set_title(f'Output mask (class {i + 1})')
            ax[i + 1].imshow(mask[1, :, :])
    else:
        ax[1].set_title(f'Output mask')
        ax[1].imshow(mask)
    plt.xticks([]), plt.yticks([])
    plt.show()


# for binary_tree.py
def unique_counts(labels):
    results = {}
    for label in labels:
        value = label.item()
        if value not in results.keys():
            results.update({value : 0})
        
        results[value] += 1
    
    return results

def log2(x):
    return log(x) / log(2)
    
# H(labels) with base 2
def entropy(labels):
    results = unique_counts(labels)
    ent = 0.0
    for r in results.keys():
        p = float(results[r]) / len(labels)
        ent = ent - p * log2(p)
    return ent

def split_function(vector, column, value):
    """
    Split function
    """
    return vector[column] >= value


def divide_set(X, y, column, value):
    """
    Divide the sets into two different sets along a specific dimension and value.
    """
    set_1 = [(vector, label) for vector, label in zip(X, y) if split_function(vector, column, value)]
    set_2 = [(vector, label) for vector, label in zip(X, y) if not split_function(vector, column, value)]

    X_set_1 = [element[0] for element in set_1]
    X_set_2 = [element[0] for element in set_2]
    y_set_1 = [element[1] for element in set_1]
    y_set_2 = [element[1] for element in set_2]

    return X_set_1, y_set_1, X_set_2, y_set_2



from multiprocessing.sharedctypes import Value
from astrape.models.models_lightning import *
from astrape.constants.astrape_constants import *
import numpy as np
from typing import Union, List, Dict, Any, Optional
import pytorch_lightning as pl
from sklearn.base import BaseEstimator
from matplotlib import pyplot as plt
import pandas as pd
from astrape.experiment import Experiment
from astrape.models.models_lightning import *
from astrape.constants.astrape_constants import *
import numpy as np
from typing import Union, List, Dict, Any, Optional, cast, overload
import pytorch_lightning as pl
from sklearn.base import BaseEstimator
import os, json
import inspect
import math
import random
from datetime import datetime
from torch.utils.tensorboard import SummaryWriter
from mpl_toolkits.mplot3d import Axes3D
import matplotlib.colors as mcolors
from sklearn.decomposition import PCA
import logging
from itertools import product
log = logging.getLogger(__name__)

available_domain_types = ["points", "image", "audio"]


class BaseProject:
    r"""Parent class of ``Project`` class. 
    
    Methods will perform 1)plotting data, 2)generating ``Experiments``,
    3)setting models across the generated `Experiment`s, 4)setting trainers across the generated ``Experiments``,
    5)training models across the generated `Experiment`s, 6)updating project metadata,
    7)adding individual ``Experiment`` to the project, 8)saving checkpoints for ``Experiments`` saved in ``exps``,
    9)saving stacks for each ``Experiment`` in ``exps``, 9)flushing all ``exps``. 

    Attributes:

        project_name (str): Name of the project. Directory of {path}/{project_name} will be created.

        X (np.ndarray or torch.Tensor or pd.DataFrame): Data for the project. If X_test is not specified, X would be splitted(stratify=y) into train/val data with a given random state.

        y (np.ndarray or torch.Tensor or pd.DataFrame): Labels for the project. If y_test is not specified, y would be splitted(stratify=y) into train/val labels with a given random state.

        X_test (np.ndarray or torch.Tensor or pd.DataFrame, optional): Test data for the project. Specify this when there is a separate test data.
            Default: ``None``

        y_test (np.ndarray or torch.Tensor or pd.DataFrame, optional): Test labels for the project. Specify this when there is a separate test data.
            Default: ``None``

        n_classes (int): Number of classes for the classification task. Don't specify this when performing regression tasks.
            Default: ``None``

        path (str): Base path for the project. Default is the current directory.
            Default: ``'.'``

        dims (:obj:`int` or :obj:`tuple` of `int`) : Dimensions of the data. Number of samples is excluded in dims.

        project_path (:obj:`str`): Path for the project. It will be {path}/{project_name}.

        log_dir (:obj:`str`): Path for saving plotted images and figures. It will be {project_path}/{projectlogs}.

        project_logger (:obj:`SummaryWriter`): TensorBoard Logger for logging.

        project_metadata (:obj:`dict`): Dictionary describing the project.

        birthday (str): Time when the project is created.

        exps (:obj:`dict` of :obj:`str`, :obj:`Experiment`): `Project` can control ``Experiments`` in ``exps``.
        
        n_exps (int): Number of ``Experiemnt``s in ``exps``.

        pseudoexp (`Experiment`): Dummy ``Experiment`` for applying methods in ``Experiment``.
    """

    def __init__(
        self,
        project_name : str,
        X : Union["np.ndarray", "torch.Tensor", "pd.DataFrame"],
        y : Union["np.ndarray", "torch.Tensor", "pd.DataFrame"],
        X_test : Optional[Union["np.ndarray", "torch.Tensor", "pd.DataFrame"]] = None,
        y_test : Optional[Union["np.ndarray", "torch.Tensor", "pd.DataFrame"]] = None,
        n_classes : Optional[int] = None,
        path : str = "."
    )->None:
        r"""Creates a ``Project``.

        Args:
            
            project_name (str): Name of the project. Directory of {path}/{project_name} will be created.

            X (np.ndarray or torch.Tensor or pd.DataFrame): Data for the project. If X_test is not specified, X would be splitted(stratify=y) into train/val data with a given random state.

            y (np.ndarray or torch.Tensor or pd.DataFrame): Labels for the project. If y_test is not specified, y would be splitted(stratify=y) into train/val labels with a given random state.

            X_test (np.ndarray or torch.Tensor or pd.DataFrame, optional): Test data for the project. Specify this when there is a separate test data.
                Default: ``None``

            y_test (np.ndarray or torch.Tensor or pd.DataFrame, optional): Test labels for the project. Specify this when there is a separate test data.
                Default: ``None``

            n_classes (int): Number of classes for the classification task. Don't specify this when performing regression tasks.
                Default: ``None``

            path (str): Base path for the project. Default is the current directory.
                Default: ``'.'``

        Returns:

            None
        """
        self.project_name = project_name
        self.X = X
        self.y = y
        self.X_test = X_test
        self.y_test = y_test
        self.n_classes = n_classes
        self.path = path
        if self.path[-1] == "/":
            self.path = self.path[:-1] # format path
        
        if len(X.shape) == 2:
            self.dims = X.shape[1]
        elif len(X.shape) > 2:
            self.dims = X.shape[1:]
        else:
            raise ValueError(f"Wrong dimension for input X.")

        self.project_path = f'{self.path}/{self.project_name}' 
        self._create_folder(self.project_path)
        self.log_dir = f'{self.project_path}/projectlogs'
        self._create_folder(self.log_dir)
        self.project_logger = SummaryWriter(log_dir=self.log_dir)

        now = datetime.now()
        birthday = now.strftime("%b-%d-%Y-%H-%M-%S")
        self.project_metadata = {'date of birth of this project' : str(birthday)}
        
        
        self.exps = {}
        self.n_exp = 0

        self._create_folder(f'{self.project_path}/results')
        for exp_name, exp in self.create_experiments(amount=1, stack_exps=False).items():
            pseudoexp = exp
        self.pseudoexp = pseudoexp # use this dummy experiement for calling Experiment methods

    def update_project_metadata(
        self
    )->None:
        r"""Updates metadata of the project.
        """
        self.project_metadata.update({'number of Experiments created' : self.n_exp})
        self.project_metadata.update({'size of this Project in MB' : float(os.path.getsize(self.project_path)/10**6)})

        json_name = self.project_path +"/project_metadata.json" 
        with open(json_name, 'w') as fout:
            json.dump(self.project_metadata, fout)
    
    def create_experiments(
        self,
        amount : int = 10,
        test_size : float = 0.01,
        stack_models : bool = True,
        stack_exps : bool = True
    )->Dict[str, "Experiment"]: # {name of the experiment} : Experiment
        r"""Generates ``Experiments`` and stack them to ``exps`` if needed.

        Args:
            
            amount (int) : The number of ``Experiments`` you want to generate.
                Default: ``10``
            
            test_size (float) : Test size when splitting the data. Neccessary for creating ``Experiments``.
                Default: ``1e-2``
            
            stack_models (bool) : Whehter to stack models in each ``Experiment`` or not.
                Default: ``True``
            
            stack_exps (bool) : Whether to stack generated ``Experiments`` to ``exps`` or not.
                Default: ``True``
        
        Returns:
            
            dict: dictionary of ``Experiments``. Will be ``exps`` if ``stack_exps`` is True, dictionary of ``Experiments`` if ``stack_exps`` is False.
        
        """
        exp_common_parameters = {
            'project_name' : self.project_name,
            'X' : self.X,
            'y' : self.y,
            'X_test' : self.X_test,
            'y_test' : self.y_test,
            'n_classes' : self.n_classes,
            'X_test' : self.X_test,
            'y_test' : self.y_test,
            'test_size' : test_size,
            'stack_models' : stack_models,
            'path' : self.path
        }
        
        temp_exps = {}

        for i in range(amount):
            seed = np.random.choice(range(10**5))
            if seed in self.random_states:
                while seed in self.random_states:
                    seed = np.random.choice(range(10**5))
            self.random_states.append(seed)
            exp_parameters = {}
            exp_parameters.update(exp_common_parameters)
            exp_parameters.update({'random_number' : seed})
            exp = Experiment(**exp_parameters)
            exp_name = f'random_state-{exp.random_state}'
            if stack_exps:
                self.exps.update({exp_name : exp})
            if not stack_exps:
                temp_exps.update({exp_name : exp})
        self.update_project_metadata()
        if stack_exps:
            return self.exps
        else:
            return temp_exps


    def add_experiment(
        self,
        experiment : Experiment
    )->None:
        r"""Add an individual ``Experiment`` to ``exps``.

        Args:

            experiment (:obj:``Experiment``): ``Experiment`` you want to add.

        Returns:

            None
        """
        random_state = experiment.random_state
        if random_state in self.random_states:
            raise ValueError('There is already an "Experiment" with the same random state.')
        exp_name = f'random_state-{random_state}'
        self.exps.update({exp_name : experiment})

        self.update_project_metadata()
    
    def set_models(
        self,
        model_type,
        random_states : Optional[List[int]] = None, # if None, same models are set for all experiments created in the Project
        **hparams
    )->None:
        r"""Set identical models across ``Experiments``.

        Args:

            model_type (``LightningModule`` or ``BaseEstimator``): Class of the model you want to use.

            **hparams (Any): Hyperparameters for the model. 

            random_states (:obj:``list`` of `int`, optional): The list of random states of ``Experiments`` you want to set model to. If None, will set models for all ``Experiments`` in ``exps``.
                Default: ``None``

        Returns:

            None
        """
        random_state_str_list = []
        
        if not random_states:
            for exp_name, exp in self.exps.items():
                exp.set_model(model_type, **hparams)
        else:
            for random_state in random_states:
                random_state_str_list.append(f'random_state-{random_state}')
            for random_state_str in random_state_str_list: 
                if random_state_str not in self.exps.keys():
                    raise ValueError(f'No "Experiment" with {random_state_str} is defined.')
                    
                exp = self.exps[random_state_str]
                exp.set_model(model_type, **hparams)
        self.update_project_metadata()

    def set_trainers(
        self,
        random_states : Optional[List[int]] = None, # if None, same trainers are set for all experiments created in the Project
        **trainer_config
    )->None:
        r"""Set identical trainers across ``Experiments``.

        Args:

            **trainer_config (Any): Trainer configurations.

            random_states (:obj:``list`` of `int`, optional): The list of random states of ``Experiments`` you want to set trainer to. If None, will set trainers for all ``Experiments`` in ``exps``.
                Default: ``None``

        Returns:
            
            None
    
        """
        random_state_str_list = []
        
        if not random_states:
            for exp_name, exp in self.exps.items():
                exp.set_trainer(**trainer_config)
        else:
            for random_state in random_states:
                random_state_str_list.append(f'random_state-{random_state}')
            for random_state_str in random_state_str_list: 
                if random_state_str not in self.exps.keys():
                    raise ValueError(f'No "Experiment" with {random_state_str} is defined.')
                exp = self.exps[random_state_str]
                exp.set_trainer(**trainer_config)
        self.update_project_metadata()

    def fit_experiments(
        self,
        random_states : Optional[List[int]] = None, # if None, all experiments will be fitted,
        **trainer_config
    )->None:
        r"""Fit models in ``Experiments`` stacked in ``exps``.

        Args:

            **trainer_config (Any): Trainer configurations.

            random_states (:obj:``list`` of `int`, optional): The list of random states of ``Experiments`` you want to fit. If None, will fit all ``Experiments`` in ``exps``.
                Default: ``None``

        Returns:

            None
        
        """
        random_state_str_list = []
        
        if not random_states:
            for exp_name, exp in self.exps.items():
                exp.fit(**trainer_config)
        else:
            for random_state in random_states:
                random_state_str_list.append(f'random_state-{random_state}')
            for random_state_str in random_state_str_list: 
                if random_state_str not in self.exps.keys():
                    raise ValueError(f'No "Experiment" with {random_state_str} is defined.')
                exp = self.exps[random_state_str]
                exp.fit(**trainer_config)
        self.update_project_metadata()

    def save_ckpts(
        self,
        random_states : Optional[List[int]] = None, # if None, all experiments will save checkpoints
    )->None:
        r"""Save checkpoints in ``Experiments`` stacked in ``exps``.

        Args:

            random_states (:obj:``list`` of `int`, optional): The list of random states of ``Experiments`` you want to save checkpoint. If None, will save checkpoint for all ``Experiments`` in ``exps``.
                Default: ``None``

        Returns:

            None
        
        """
        random_state_str_list = []
        if not random_states:
            for exp_name, exp in self.exps.items():
                exp.save_ckpt()
        else:
            for random_state in random_states:
                random_state_str_list.append(f'random_state-{random_state}')
                for random_state_str in random_state_str_list: 
                    if random_state_str not in self.exps.keys():
                        raise ValueError(f'No "Experiment" with {random_state_str} is defined.')
                    exp = self.exps[random_state_str]
                    exp.save_ckpt()
        self.update_project_metadata()

    def save_stacks(
        self,
        random_states : Optional[List[int]] = None, # if None, all experiments will save stacks
    )->None:
        r"""Apply ``save_stack()`` for ``Experiments`` in ``exps``.

        Args:

            random_states (:obj:``list`` of `int`, optional): The list of random states of ``Experiments`` you want to save stack. If None, will save stack for all ``Experiments`` in ``exps``.
                Default: ``None``

        Returns:

            None
        
        """
        random_state_str_list = []
        if not random_states:
            for exp_name, exp in self.exps.items():
                exp.save_stack()
        else:
            for random_state in random_states:
                random_state_str_list.append('random_state-{random_state}')
                for random_state_str in random_state_str_list: 
                    if random_state_str not in self.exps.keys():
                        raise ValueError(f'No "Experiment" with {random_state_str} is defined.')
                    exp = self.exps[random_state_str]
                    exp.save_stack()
        self.update_project_metadata()
        
    def save_project(
        self
    )->None:
        r"""Save all ``Experiments``.
        """
        if self.exps:
            self.save_stacks()
        self.update_project_metadata()

    def flush_exps(
        self,
        save_stacks : bool = True
    )->None:
        r"""Flushes ``exps`` after saving it.
        """
        if self.exps and save_stacks:
            self.save_stacks()
        del self.exps
        self.exps = {}
        self.update_project_metadata()

    def plot_data(
        self,
        domain_type : str,
        dataformats : Optional[str] = "NHWC",
        n_data : Optional[int] = 10
    )->None:
        r"""Visualizes the data.
        
        Args:

            domain_type (str): The domain type the data inhabits in. e.g., "image", "points"

            dataformats (str, optional): The dimensions specifying the orderings of (`N`,`H`,`W`,`C`) for image data. Does nothing when domain type is not "image".
                Default: ``NHWC``

            n_data (int): Number of image samples you want to visualize. Does nothing when domain type is not "image".
                Default: ``10``
        
        Returns:

            None
        """
        points_dir = self.project_path + "/visualizations"
        self._create_folder(points_dir)
        if domain_type == "points": # data points
            if self.dims == 1: # 1D
                fig = plt.figure()
                points_dict = {}
                for label_idx in range(self.n_classes):
                    points_dict.update({label_idx: self.X[self.y==label_idx]})
                    plt.scatter(x=points_dict[label_idx][:,0], y=points_dict[label_idx][:,1], label=label_idx)
                plt.title('Data Distribution')
                plt.xlabel('positions')
                plt.legend(loc=(1.05,1.05))
                plt.show()
                plt.savefig(points_dir+ "/points_distribution.png")  
                self.project_logger.add_figure('data distribution', fig)
            elif self.dims == 2 : # 2D
                fig = plt.figure()
                points_dict = {}
                for label_idx in range(self.n_classes):
                    points_dict.update({label_idx: self.X[self.y==label_idx]})
                    plt.scatter(x=points_dict[label_idx][:,0], y=points_dict[label_idx][:,1], label=label_idx)
                plt.title('Data Distribution')
                plt.xlabel('x positions')
                plt.ylabel('y positions')
                plt.legend(loc=(1.,1.))
                plt.show()
                plt.savefig(points_dir+ "/points_distribution.png")  
                self.project_logger.add_figure('data distribution', fig)
            elif self.dims == 3 : # 3D 
                fig1 = plt.figure()
                ax = fig1.add_subplot(111, projection='3d')
                points_dict = {}
                for label_idx in range(self.n_classes):
                    points_dict.update({label_idx: self.X[self.y==label_idx]})
                    ax.scatter(xs=points_dict[label_idx][:,0], ys=points_dict[label_idx][:,1], zs= points_dict[label_idx][:,2], label=label_idx)
                ax.set_xlabel('x positions')
                ax.set_ylabel('y positions')
                ax.set_zlabel('z positions')
                plt.legend(loc=(1.,1.))
                plt.title('Data Distribution')
                plt.show()
                plt.savefig(points_dir+ "/points_distribution_1.png")  
                self.project_logger.add_figure('data_distribution_1', fig1)
                plt.close(fig1)
                fig2 = plt.figure()
                ax = fig2.add_subplot(111, projection='3d')
                points_dict = {}
                for label_idx in range(self.n_classes):
                    points_dict.update({label_idx: self.X[self.y==label_idx]})
                    ax.scatter(xs=points_dict[label_idx][:,0], ys=points_dict[label_idx][:,1], zs= points_dict[label_idx][:,2], label=label_idx)
                ax.set_xlabel('x positions')
                ax.set_ylabel('y positions')
                ax.set_zlabel('z positions')
                ax.view_init(0, 90)
                plt.legend(loc=(1.,1.))
                plt.title('Data Distribution')
                plt.show()
                plt.savefig(points_dir+ "/points_distribution_2.png")  
                self.project_logger.add_figure('data_distribution_2', fig2)
                plt.close(fig2)
                fig3 = plt.figure()
                ax = fig3.add_subplot(111, projection='3d')
                points_dict = {}
                for label_idx in range(self.n_classes):
                    points_dict.update({label_idx: self.X[self.y==label_idx]})
                    ax.scatter(xs=points_dict[label_idx][:,0], ys=points_dict[label_idx][:,1], zs= points_dict[label_idx][:,2], label=label_idx)
                ax.set_xlabel('x positions')
                ax.set_ylabel('y positions')
                ax.set_zlabel('z positions')
                ax.view_init(0, 180)
                plt.legend(loc=(1.,1.))
                plt.title('Data Distribution')
                plt.show()
                plt.savefig(points_dir+ "/points_distribution_3.png")  
                self.project_logger.add_figure('data_distribution_3', fig3)
                plt.close(fig3)
                fig4 = plt.figure()
                ax = fig4.add_subplot(111, projection='3d')
                points_dict = {}
                for label_idx in range(self.n_classes):
                    points_dict.update({label_idx: self.X[self.y==label_idx]})
                    ax.scatter(xs=points_dict[label_idx][:,0], ys=points_dict[label_idx][:,1], zs= points_dict[label_idx][:,2], label=label_idx)
                ax.set_xlabel('x positions')
                ax.set_ylabel('y positions')
                ax.set_zlabel('z positions')
                ax.view_init(0, 270)
                plt.legend(loc=(1.,1.))
                plt.title('Data Distribution')
                plt.show()
                plt.savefig(points_dir+ "/points_distribution_4.png")  
                self.project_logger.add_figure('data_distribution_4', fig4)
                plt.close(fig4)
            else: # higher dimensions -> use PCA
                log.detail(f'The dimension of data is higher than 3. (Input dimension: {self.dims})\n \
                     We will use PCA and visualize the data using 2 principal axes.')
                pca = PCA(n_components=2)
                pca_data = pca.fit_transform(self.X)
                fig = plt.figure()
                points_dict = {}
                for label_idx in range(self.n_classes):
                    points_dict.update({label_idx: pca_data[self.y==label_idx]})
                    plt.scatter(x=points_dict[label_idx][:,0], y=points_dict[label_idx][:,1], label=label_idx)
                plt.title('Data Distribution')
                plt.xlabel('1st principal axis')
                plt.ylabel('2nd principal axis')
                plt.legend(loc=(1.,1.))
                plt.show()
                plt.savefig(points_dir+ f"/points_distribution_PCA.png")  
                self.project_logger.add_figure('data distribution_PCA', fig)

                
    
        elif domain_type == "image":

            if len(self.dims) == 1 :
                raise AssertionError(f"Wrong dimension {self.dims} for image data. It should be at least 2-dimensional.")
            img_dir =  f"{self.project_path}/visualizations"
            self._create_folder(img_dir)
            
            if len(self.dims) == 2 : # 2D, black & white
                imgs = np.reshape(self.X, (-1, *self.dims, 1)) # (# of samples, height, width, # of channels) i.e., NHWC       
            elif len(self.dims) > 2 : # 2D, colored images
                imgs = np.reshape(self.X, (-1, *self.dims)) # (# of samples, height, width, # of channels) i.e., NHWC


            sampling_idx = np.random.choice(range(imgs.shape[0]), size=n_data)
            imgs_to_show = imgs[sampling_idx]
            labels_to_show = self.y[sampling_idx]
            # for logging
            self.project_logger.add_images('sample_images', imgs_to_show, dataformats=dataformats)
            dataformats_without_N = dataformats.replace("N","")
            for i in range(n_data):
                self.project_logger.add_image(
                    f"sample_images/image_{i}_label_{labels_to_show[i]}",
                    imgs_to_show[i],
                    dataformats=dataformats_without_N
                )
            # for plotting
            rows = int(math.sqrt(n_data)) + 1
            cols = int(math.sqrt(n_data))
            fig = plt.figure(figsize=(5*rows, 5*cols))
            idx = 0
            for i in range(1, rows+1):
                for j in range(1, cols+1):
                    idx += 1
                    if idx < n_data + 1:
                        img = fig.add_subplot(rows, cols, idx)
                        img.imshow(self.X[sampling_idx][idx-1])
                        img.set_title('label : {labels_to_show[idx-1]}')
            plt.show()
            plt.savefig(f"{img_dir}/image_samples.png")       

        elif self.domain_type == "audio":
            raise NotImplementedError("Audio data unimplemented.") # TODO: implement this
        else:
            raise NotImplementedError("Unsupported domain type {self.domain_type}. Supported domain types are {[*available_domain_types]}")
        
        self.update_project_metadata()

    def get_available_random_states(
        self
    )->List[str]:
        r"""Returns list of directories of ``Experiments`` that has been fitted and logged.

        Returns:

            list of `str`: All directories of fitted and logged ``Experiments``.
        """
        log_base_dir = f"{self.project_path}/FIT"
        random_state_dir_list = []
        for random_state_dir in os.listdir(log_base_dir):
            random_state_dir_list.append(f'{log_base_dir}/{random_state_dir}')

        return random_state_dir_list

    def search_standard(
        self,
        val_metric : str = 'val/acc'
    )->Tuple["np.ndarray", dict[str,int], dict[str,int]]: 
        r"""Search method for finding the model type/structure with best validation performance.

        Args:

            val_metric (str): The validation metric.
                Default: ``'val/acc'``

        Returns:

            tuple: (numpy ndarray of validation performances across ``Experiments``, dictionary of {model type:it's frequency}, dictionary of {model structure:it's frequency}).

        """ 
        performances = []
        best_model_type = None
        best_metadata = None
        best_val = None
        best_model_structures = {}
        best_model_types = {}
        mode = 'min' if val_metric == 'val/loss' else 'max'
        random_state_dir_list = self.get_available_random_states()
        for random_state_dir in random_state_dir_list:
            jsonlog_dir = f"{random_state_dir}/logs/jsonlogs"
            for model_type_dir in os.listdir(jsonlog_dir): 
                for metadata_dir in os.listdir(f'{jsonlog_dir}/{model_type_dir}'): 
                    for version in os.listdir(f'{jsonlog_dir}/{model_type_dir}/{metadata_dir}'): 
                        for file in os.listdir(f'{jsonlog_dir}/{model_type_dir}/{metadata_dir}/{version}'): 
                            file_path = f'{jsonlog_dir}/{model_type_dir}/{metadata_dir}/{version}/{file}'  
                            with open(file_path, "r") as json_file:
                                logs = json.load(json_file)
                                val_score = logs[val_metric] if isinstance(logs[val_metric], float) else logs[val_metric][-1]
                                if best_val is None:
                                    best_val = val_score
                                    best_model_type = model_type_dir
                                    best_metadata = metadata_dir
                                elif mode == "max" and val_score > best_val:
                                    best_val = val_score
                                    best_model_type = model_type_dir
                                    best_metadata = metadata_dir
                                elif mode == "min" and val_score < best_val:
                                    best_val = val_score
                                    best_model_type = model_type_dir
                                    best_metadata = metadata_dir
            # update best_model_types for each random state
            if best_model_type not in best_model_types.keys():
                best_model_types.update({best_model_type : 1})
            else:
                best_model_types.update({best_model_type : best_model_types[best_model_type]+1})

            # update best_model_structures for each random state
            if best_metadata not in best_model_structures.keys():
                best_model_structures.update({best_metadata: 1})
            else:
                best_model_structures.update({best_metadata : best_model_structures[best_metadata]+1})
            # update performances for each random state
            if best_val:
                performances.append(best_val)
        performances = np.array(performances)
        if performances == np.array([]):
            raise ValueError(f"No performances are saved.")
        return (performances, best_model_types, best_model_structures)

    def search_with_model_type(
        self,
        model_type : Union["pl.LightningModule", "BaseEstimator"],
        val_metric : str = 'val/acc'
    )->Tuple["np.ndarray",Dict[str,int]]:
        r"""Search method for finding validation performances for a specific model type.

        Args:

            model_type (`LightningModule` or `BaseEstimator`): The model type of interest. e.g., ``MLP``, ``LogisticRegression``.

            val_metric (str): The validation metric.
                Default: ``'val/acc'``

        Returns:

            tuple: (numpy ndarray of validation performances, dictionary of {model structure:it's frequency}).
        """
        performances = []
        best_model_structures = {}
        best_val = None
        best_metadata = None
        mode = 'min' if val_metric == 'val/loss' else 'max'
        random_state_dir_list = self.get_available_random_states()
        for random_state_dir in random_state_dir_list:
            jsonlog_dir = f'{random_state_dir}/logs/jsonlogs'
            for model_type_dir in os.listdir(jsonlog_dir): 
                if str(model_type_dir) == model_type.__name__:
                    for metadata_dir in os.listdir(f'{jsonlog_dir}/{model_type_dir}'): 
                        for version in os.listdir(f'{jsonlog_dir}/{model_type_dir}/{metadata_dir}'): 
                            for file in os.listdir(f'{jsonlog_dir}/{model_type_dir}/{metadata_dir}/{version}'):
                                file_path = f'{jsonlog_dir}/{model_type_dir}/{metadata_dir}/{version}/{file}'
                                with open(file_path, "r") as json_file:
                                    logs = json.load(json_file)
                                    val_score = logs[val_metric] if isinstance(logs[val_metric], float) else logs[val_metric][-1]
                                    if best_val is None:
                                        best_val = val_score
                                        best_metadata = metadata_dir
                                    elif mode == "max" and val_score > best_val:
                                        best_val = val_score
                                        best_metadata = metadata_dir
                                    elif mode == "min" and val_score < best_val:
                                        best_val = val_score
                                        best_metadata = metadata_dir
            # update performances for each random state
            if best_val:
                performances.append(best_val)
            # update best_model_structures for each random state
            if best_metadata not in best_model_structures.keys():
                best_model_structures.update({best_metadata: 1})
            else:
                best_model_structures.update({best_metadata : best_model_structures[best_metadata]+1})
            
        performances = np.array(performances)

        if performances == np.array([]):
            raise ValueError(f"None is in performances. You may have not saved {model_type}")

        return (performances, best_model_structures)

    def search_with_model_type_hparams(
        self,
        model_type : Union["pl.LightningModule", "BaseEstimator"], 
        val_metric : str = 'val/acc',
        **hparams
    )->Tuple["np.ndarray", str]:
        r"""Search method for finding validation performances of a specific model structure.

        Args:

            model_type (`LightningModule` or `BaseEstimator`): The model type of interest. e.g., ``MLP``, ``LogisticRegression``.

            **hparams (Any): Hyperparameters for the model.

            val_metric (str): The validation metric.
                Default: ``'val/acc'``

        Returns:

            tuple: (numpy ndarray of validation performances, metadata specifying the specified model structure).
        """
        performances = []
        best_val = None
        
        if BaseEstimator in inspect.getmro(model_type):
            pseudomodel = self.pseudoexp.set_model(model_type, **hparams)
            metadata = self.pseudoexp.set_model_metadata(pseudomodel, pseudomodel.__dict__)
        elif pl.LightningModule in inspect.getmro(model_type):
            pseudomodel = self.pseudoexp.set_model(model_type, **hparams)
            metadata = self.pseudoexp.set_model_metadata(pseudomodel, pseudomodel.hparams)     
        mode = 'min' if val_metric == 'val/loss' else 'max'
        random_state_dir_list = self.get_available_random_states()
        for random_state_dir in random_state_dir_list:
            jsonlog_dir = random_state_dir+"/logs/jsonlogs"
            for model_type_dir in os.listdir(jsonlog_dir):
                for metadata_dir in os.listdir(jsonlog_dir + "/" + model_type_dir):
                    if str(metadata_dir) == metadata: 
                        best_val = None
                        for version in os.listdir(jsonlog_dir + "/" + model_type_dir + "/" + metadata_dir):
                            for file in os.listdir(jsonlog_dir + "/" + model_type_dir + "/" + metadata_dir + "/" + version):
                                file_path = jsonlog_dir + "/" + model_type_dir + "/" + metadata_dir + "/" + version + "/" + file
                                with open(file_path, "r") as json_file:
                                    logs = json.load(json_file)
                                    val_score = logs[val_metric] if isinstance(logs[val_metric], float) else logs[val_metric][-1]
                                    if best_val is None:
                                        best_val = val_score
                                    elif mode == "max" and val_score > best_val:
                                        best_val = val_score
                                    elif mode == "min" and val_score < best_val:
                                        best_val = val_score
            if best_val:
                performances.append(best_val)
            
        performances = np.array(performances)
        if performances == np.array([]):
            raise ValueError(f"None is in performances. You may have inputted the wrong model strucutre.")
        return (performances, metadata)
                    
    def search_all_models(
        self,
        val_metric : str = 'val/acc'
    )->Tuple["np.ndarray", List[str]]:
        r"""Method for searching validation performances of all saved models.

        Args:

            val_metric (str): The validation metric.
                Default: ``'val/acc'``

        Returns:

            tuple: (numpy ndarray of list of validation performances, list of searched model metadata)
        """
        metadata_list = []
        performances_list = [] # list of performances(list). len(performances_list) == len(metadata_list)
        mode = 'min' if val_metric == 'val/loss' else 'max'
        random_state_dir_list = self.get_available_random_states()
        random_number_list = []
        best_val = None
        for random_number_dir in random_state_dir_list:
            random_number = random_number_dir.split("-")[-1]
            random_number_list.append(random_number)

        performances_per_metadata = {}
        for random_state_dir in random_state_dir_list:
            jsonlog_dir = f'{random_state_dir}/logs/jsonlogs'
            for model_type_dir in os.listdir(jsonlog_dir): 
                for metadata_dir in os.listdir(f'{jsonlog_dir}/{model_type_dir}'):
                    if metadata_dir not in metadata_list:
                        metadata_list.append(metadata_dir) 
                        performances_per_metadata.update({metadata_dir: []})
                    best_val = None
                    for version in os.listdir(f'{jsonlog_dir}/{model_type_dir}/{metadata_dir}'): 
                        for file in os.listdir(f'{jsonlog_dir}/{model_type_dir}/{metadata_dir}/{version}'): 
                            file_path = f'{jsonlog_dir}/{model_type_dir}/{metadata_dir}/{version}/{file}' 
                            with open(file_path, "r") as json_file:
                                logs = json.load(json_file)
                                val_score = logs[val_metric] if isinstance(logs[val_metric], float) else logs[val_metric][-1]
                                if best_val is None:
                                    best_val = val_score
                                elif mode == "max" and val_score > best_val:
                                    best_val = val_score
                                elif mode == "min" and val_score < best_val:
                                    best_val = val_score
                    if best_val:
                        performances_per_metadata[metadata_dir].append(best_val)
        for metadata in metadata_list:
            performances_list.append(np.array(performances_per_metadata[metadata]))
        performances_list = np.array(performances_list)

        if not metadata_list or None in performances_list or performances_list == np.array([]):
            raise ValueError(f"Nothing has been logged.") 
        return (metadata_list, performances_list)

    @staticmethod
    def _create_folder(directory : str)->str: #path of directory
        r"""Creates folder.
        
        Args:

            directory (str): The path of folder you want to create. When the folder already exists, the method does nothing.

        Returns: 

            str: directory 
        """
        os.makedirs(directory, exist_ok=True)
        return directory

    
        

class Project(BaseProject):
    r"""Project class that plots results of the ML project.

    Attributes:

        project_name: Name of the project. Directory of {path}/{project_name} will be created.

        X: Data for the project. If X_test is not specified, X would be splitted(stratify=y) into train/val data with a given random state.

        y: Labels for the project. If y_test is not specified, y would be splitted(stratify=y) into train/val labels with a given random state.

        X_test : Test data for the project. Specify this when there is a separate test data.
            Default: ``None``

        y_test : Test labels for the project. Specify this when there is a separate test data.
            Default: ``None``

        n_classes : Number of classes for the classification task. Don't specify this when performing regression tasks.
            Default: ``None``

        path : Base path for the project. Default is the current directory.
            Default: ``.``

        dims (:obj:`int` or :obj:`tuple` of `int`) : Dimensions of the data. Number of samples is excluded in dims.

        project_path (:obj:`str`): Path for the project. It will be {path}/{project_name}.

        log_dir (:obj:`str`): Path for saving plotted images and figures. It will be {project_path}/{projectlogs}.

        project_logger (:obj:`SummaryWriter`): TensorBoard Logger for logging.

        project_metadata (:obj:`dict`): Dictionary describing the project.

        birthday (str): Time when the project is created.

        exps (:obj:`dict` of :obj:`str`, :obj:`Experiment`): `Project` can control ``Experiments`` in ``exps``.
        
        n_exps (int): Number of ``Experiemnt``s in ``exps``.

        pseudoexp (`Experiment`): Dummy ``Experiment`` for applying methods in ``Experiment``.

    Methods:


    """
    def __init__(
        self,
        project_name : str,
        X : "np.ndarray",
        y : "np.ndarray",
        X_test : Optional["np.ndarray"] = None,
        y_test : Optional["np.ndarray"] = None,
        n_classes : Optional[int] = None,
        path : str = "."
    )->None:
        r"""
        
        Args:

            project_name: Name of the project. Directory of {path}/{project_name} will be created.

            X: Data for the project. If X_test is not specified, X would be splitted(stratify=y) into train/val data with a given random state.

            y: Labels for the project. If y_test is not specified, y would be splitted(stratify=y) into train/val labels with a given random state.

            X_test : Test data for the project. Specify this when there is a separate test data.
                Default: ``None``

            y_test : Test labels for the project. Specify this when there is a separate test data.
                Default: ``None``

            n_classes : Number of classes for the classification task. Don't specify this when performing regression tasks.
                Default: ``None``

            path : Base path for the project. Default is the current directory.
                Default: ``.``        
        """
        super().__init__(
            project_name=project_name,
            X=X,
            y=y,
            X_test=X_test,
            y_test=y_test,
            n_classes=n_classes,
            path=path
        )

    def plot_best_performances(
            self,
            val_metric : str = 'val/acc',
            plt_title : Optional[str] = None
        )->Tuple["np.ndarray",Dict[str,Dict[str,int]]]: # (validation performances, information)
            r"""Plots the best validation performance for each ``Experiment``.

            Args:

                val_metric (str): The validation metric.
                    Default: ``'val/acc'``

                plt_title (str, optional): The title for the plot. If left None, title will be 'Best Performance for Each Random State'.
                    Default: None

            Returns:

                tuple: (numpy ndarray of validation performances, dictionary containing information of the search)
            """
            performances, best_model_types, best_model_structures = self.search_standard(val_metric)
            random_state_dir_list = self.get_available_random_states()
            ymin = performances.min() - 0.1 if performances.min() > 0.1 else 0
            ymax = performances.max() + 0.1 if performances.max() < 0.9 else 1
            random_number_list = []
            for random_number_dir in random_state_dir_list:
                random_number = random_number_dir.split("-")[-1]
                random_number_list.append(random_number)
            fig1 = plt.figure()
            x = np.arange(len(performances)) / 2
            plt.bar(x, performances, width=0.1)
            plt.xticks(x, random_number_list)
            plt.xticks(rotation=70)
            plt.xlabel("random state")
            plt.grid(True)
            plt.ylabel(val_metric)

            if not plt_title:
                plt_title = 'Best Performance for Each Random State'
            plt.title(plt_title)
            plt.ylim = (ymin, ymax)
            self._create_folder(self.project_path+"/results")
            plt.show()
            plt.savefig(self.project_path+f"/results/{plt_title}.png")
            self.project_logger.add_figure(f"results/{plt_title}", fig1)
            
            plt.close(fig1)

            info = {'best model structures' : best_model_structures, 'best model types' : best_model_types}

            return (performances, info)

    def plot_best_model_frequencies(
        self,
        val_metric : str = 'val/acc',
        plt_title : Optional[str] = None
    )->Tuple["np.ndarray",Dict[str,Dict[str,int]]]: # (validation performances, information)
        r"""Plots pie graphs showing the frequencies of 1)best model type and 2)best model structure for each ``Experiment``.

        Args:

            val_metric (str): The validation metric/
                Default: ``'val/acc'``
            
            plt_title (str, optional): The title for the plot. If left None, title will be 'Frequencies of the Best Model Types/Structures' respectively.
                Default: None

            Returns:

                tuple: (numpy ndarray of validation performances, dictionary containing information of the search)
        """
        performances, best_model_types, best_model_structures = self.search_standard(val_metric)
        fig2 = plt.figure()
        sizes = list(best_model_structures.values())
        wedges, texts, autotexts = plt.pie(sizes, autopct='%1.1f%%',textprops=dict(color="w"))
        if not plt_title:
            plt_title = 'Frequencies of'
        plt.title(f'{plt_title} the Best Model Structures')
        plt.legend(wedges, best_model_structures.keys(), title="Model Structures", loc="center left",bbox_to_anchor=(1, 0, 0.5, 1))
        plt.setp(autotexts, size=8, weight="bold")
        plt.axis('equal')
        plt.show()
        plt.savefig(self.project_path+f"/results/{plt_title} the Best Model Structures.png")
        self.project_logger.add_figure(f"results/{plt_title} the Best Model Structures", fig2)
        plt.close(fig2)
        
        fig3 = plt.figure()
        sizes = list(best_model_types.values())
        wedges, texts, autotexts = plt.pie(sizes,  autopct='%1.1f%%') 
        plt.axis('equal')
        plt.setp(autotexts, size=8, weight="bold")
        plt.title(f'{plt_title} the Best Model Types')
        plt.legend(wedges, best_model_types.keys(), title="Model Types", loc="center left",bbox_to_anchor=(1, 0, 0.5, 1))
        plt.show()
        plt.savefig(self.project_path+f"/results/{plt_title} the Best Model Types.png")
        self.project_logger.add_figure(f"results/{plt_title} the Best Model Types", fig3)
        plt.close(fig3)

        self.update_project_metadata()
        info = {'best model structures' : best_model_structures, 'best model types' : best_model_types}

        return (performances, info)
        

    def plot_identical_model_type(
        self,
        model_type : Union["pl.LightningModule","BaseEstimator"],
        val_metric : str = "val/acc",
        plt_title : Optional[str] = None
    )->Tuple["np.ndarray", List[str], Dict[str,Dict[str,int]]]: 
        r"""Plots validation performances across different random states for a given model type.

        Args:

            model_type (`LightningModule` or `BaseEstimator`): The model type of interest.

            val_metric (str): The validation metric.
                Default: ``'val/acc'``

            plt_title (str, optional): The title for the plot. If left None, title will be 'Best Performances : Model Type {name of the model type}' respectively.
                Default: ``None``

        Returns:

            tuple: (numpy ndarray of validation performances, list of random states of saved ``Experiments``, dictionary containing information of the search)
        """
        performances, best_model_structures = self.search_with_model_type(val_metric=val_metric, model_type=model_type)
        random_state_dir_list = self.get_available_random_states()
        ymin = performances.min() - 0.1 if performances.min() > 0.1 else 0
        ymax = performances.max() + 0.1 if performances.max() < 0.9 else 1
        random_number_list = []
        for random_number_dir in random_state_dir_list:
            random_number = random_number_dir.split("-")[-1]
            random_number_list.append(random_number)

        fig1 = plt.figure()
        x = np.arange(len(random_number_list))
        plt.bar(x, performances, width=0.1)
        plt.xticks(x, random_number_list)
        plt.xticks(rotation=70)
        plt.xlabel("random state")
        plt.ylabel(val_metric)
        plt.grid(True)
        plt.ylim = (ymin, ymax)
        if not plt_title:
            plt_title = f'Best Performances : Model Type \n{model_type.__name__}'
        plt.title(plt_title)
        plt.show()
        plt.savefig(self.project_path+f"/results/{plt_title}.png")
        self.project_logger.add_figure(f"results/{plt_title}", fig1)
        plt.close(fig1)
        self.update_project_metadata()
        info = {'best model structures' : best_model_structures}

        return (performances, random_number_list, info)

    def plot_identical_model_structure(
        self,
        model_type : Union["pl.LightningModule","BaseEstimator"],
        val_metric : str = "val/acc",
        plt_title : Optional[str] = None,
        **hparams
    )->Tuple["np.ndarray",List[str]]: 
        r"""Plots validation performances across different random states for a given model structure.

        Args:

            model_type (`LightningModule` or `BaseEstimator`): The model type of interest.

            **hparams (Any): Model hyperparameters. 

            val_metric (str): The validation metric.
                Default: ``'val/acc'``

            plt_title (str, optional): The title for the plot. If left None, title will be 'Performances of a Model Structure'.
                Default: ``None``

        Returns:

            tuple: (numpy ndarray of validation performances, list of random states of saved ``Experiments``)
        """
        performances, metadata = self.search_with_model_type_hparams(
            val_metric=val_metric, model_type=model_type, **hparams
            )
        random_state_dir_list = self.get_available_random_states()
        ymin = performances.min() - 0.1 if performances.min() > 0.1 else 0
        ymax = performances.max() + 0.1 if performances.max() < 0.9 else 1
        random_number_list = []
        for random_number_dir in random_state_dir_list:
            random_number = random_number_dir.split("-")[-1]
            random_number_list.append(random_number)
        fig1 = plt.figure()
        x = np.arange(len(performances))
        plt.plot(performances, 'r.')
        plt.plot(performances, label=metadata)
        plt.xticks(x, random_number_list)
        plt.xticks(rotation=70)
        plt.xlabel("random state")
        plt.ylabel(val_metric)
        plt.grid(True)
        plt.legend(title="Model Structures", loc="center left",bbox_to_anchor=(1, 0, 0.5, 1))
        plt.ylim = (ymin, ymax)
        
        if not plt_title:
            plt_title = 'Performances of a Model Structure'
        plt.title(plt_title)
        plt.savefig(self.project_path+f"/results/{metadata}.png")
        plt.show()
        self.project_logger.add_figure(f'results/{metadata}', fig1)
        plt.close(fig1)

        self.update_project_metadata()

        return (performances, random_number_list)
    
    @staticmethod
    def generate_marker_style(amount)->List[str]: 
        r"""Returns a string specifying the marker style
        """
        color_list = ['b','g','r','c','m','y','k','w']
        shape_list = [".", "v", "1", "*", "^", "s", "<"]
        random.shuffle(color_list)
        random.shuffle(shape_list)
        marker_styles = []
        for color, shape in product(color_list, shape_list):
            marker_styles.append(color+shape)
            if len(marker_styles) == amount:
                break
        return marker_styles

    def plot_all_model_structures(
        self,
        val_metric : str = "val/acc",
        plt_title : Optional[str] = None
    )->Tuple["np.ndarray", List[str]]:
        r"""Plots validation performances across different random states for all model structures saved. 
        
        Will plot 1)line graphs and 2)box plots.

        Args:

            val_metric (str): The validation metric.
                Default: ``'val/acc'``

            plt_title (str, optional): The title for the plot. If left None, title will be 'Performances For All Models' and 'Performances For All Models:Box Plot'.
                Default: ``None``

        Returns:

            tuple: (numpy ndarray of validation performances, list of random states of saved ``Experiments``)
        """
    # (validation performances, list of model metadata)
        
        metadata_list, performances_list = self.search_all_models(val_metric)
        random_state_dir_list = self.get_available_random_states()
        random_number_list = []
        for random_number_dir in random_state_dir_list:
            random_number = random_number_dir.split("-")[-1]
            random_number_list.append(random_number)
        fig1 = plt.figure()
        ymin_list = ymax_list = []
        marker_styles = self.generate_marker_style(len(metadata_list))
        for idx, (metadata, performance_per_metadata) in enumerate(zip(metadata_list, performances_list)):
            ymin_list.append(performance_per_metadata.min())
            ymax_list.append(performance_per_metadata.max())
            plt.plot(performance_per_metadata, marker_styles[idx])
            plt.plot(performance_per_metadata, label=metadata)
            
        ymin_list, ymax_list = np.array(ymin_list), np.array(ymax_list)
        ymin, ymax = ymin_list.min(), ymax_list.max()
        
        if not plt_title:
            plt_title = 'Performances For All Models'
        plt.title(plt_title)
        x = np.arange(len(random_number_list))
        plt.xticks(x, random_number_list)
        plt.xticks(rotation=70)
        plt.xlabel("random state")
        plt.ylabel(val_metric)
        plt.grid(True)
        plt.ylim = (ymin, ymax)
        plt.legend(title="Model Structures", loc="center left",bbox_to_anchor=(1, 0, 0.5, 1))
        plt.show()
        plt.savefig(self.project_path+f"/results/{plt_title}.png")
        self.project_logger.add_figure(f'results/{plt_title}', fig1)
        
        plt.close(fig1)

        fig2 = plt.figure()
        plot_dict = {}
        for metadata, performances in zip(metadata_list, performances_list):    
            plot_dict.update({metadata : performances}) 
        keys = list(plot_dict.keys())
        data = list(plot_dict.values())
        
        legend_dict = {}
        for idx, metadata in enumerate(keys):
            bp = plt.boxplot(
                data[idx], positions=[idx+1], 
                patch_artist=True, boxprops=dict(facecolor=f"C{idx}")
                )
            legend_dict.update({f'bp{idx}': bp})
        plt.legend(
            [bp["boxes"][0] for bp in legend_dict.values()], 
            [metadata for metadata in keys],
            title="Model Structures", 
            loc="center left", 
            bbox_to_anchor=(1, 0, 0.5, 1)
        )
    
        plt.title(f'{plt_title}: Box Plot')
        plt.xlabel(f"model structure")
        plt.ylabel(val_metric)
        plt.grid(True)
        plt.show()
        plt.savefig(self.project_path+f"/results/{plt_title}_boxplot.png")
        self.project_logger.add_figure(f'results/{plt_title}_boxplot', fig2)

        return (performances_list, metadata_list)

    




from lib2to3.pytree import Base
from pyexpat.errors import XML_ERROR_INVALID_TOKEN

import pytorch_lightning as pl
import numpy as np
from sklearn.base import BaseEstimator

from astrape.base.experiment_base import BaseExperiment
from astrape.utilities.utils_lightning import *
from astrape.exceptions.exceptions import *
from astrape.model_selection import CrossValidation
from astrape.constants.astrape_constants import *

from typing import Union, List, Dict, Any, Optional, Tuple, cast, overload

import warnings
warnings.filterwarnings(action='ignore')

class Experiment(BaseExperiment):
    r"""``Experiment`` class.

    Note:
        This class is just ``BaseExperiment`` with ``cross_validation``.
        See ``astrape.base.experiment_base`` for details.
    """
    def __init__(
        self,
        project_name : str,
        X : "np.ndarray",
        y : "np.ndarray",
        n_classes : Optional[int] = None,
        X_test : Optional[np.ndarray] = None,
        y_test : Optional[np.ndarray] = None,
        test_size : float = 0.01,
        stack_models : bool = True,
        random_number : int = 0,
        path : Optional[str] = None
    )->None:
       
        super().__init__(
            project_name=project_name,
            X=X,
            y=y,
            n_classes=n_classes,
            X_test=X_test,
            y_test=y_test,
            test_size=test_size,
            stack_models=stack_models,
            random_number=random_number,
            path=path
        )
    def cross_validation(
        self,
        model_type : Union["BaseEstimator", "pl.LightningModule"],
        parameters : Dict[str, List[Any]],
        search_type : str = "random",
        val_metric : str = 'val/acc',
        mode : Optional[str] = None,
        trainer_config : Optional[Dict] = None,
        n_random_search : Optional[int] = None,
        cv : int = 5,
        random_number : int = DEFAULT_RANDOM_SEED_NUMBER 
    )->Tuple[Union["pl.LightningModule", "BaseEstimator"],Dict]:
        self.fit_or_cv = "CV"
        self.update_log_path()
        best_model_structure, info = CrossValidation(
            project_name=self.project_name,
            dims=self.dims,
            X=self.X,
            y=self.y,
            n_classes=self.n_classes,
            X_test=self.X_test,
            y_test=self.y_test,
            test_size=self.test_size,
            stack_models=self.stack_models,
            path=self.path,
            model_type=model_type,
            parameters=parameters,
            trainer_config=trainer_config,
            val_metric=val_metric,
            mode=mode,
            cv=cv,
            search_type=search_type,
            n_random_search=n_random_search,
            random_number=random_number
        ).search()
        self.fit_or_cv = "FIT"
        return (best_model_structure, info)
        
    